"""Heatmap loss functions."""

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from landmarker.models.utils import LogSoftmaxND


class GeneralizedNormalHeatmapLoss(nn.Module):
    """
    Loss function for adaptive generalized normal direct heatmap regression. The loss function is an
    extension of the loss function proposed by Thaler et al. (2021) for adaptive heatmap regression,
    where they used a anistropic Gaussian distribution for adaptive heatmap regression. The loss
    function is defined as the sum of a specified distance function between the predicted heatmap
    and the target heatmap, additionaly a determinant of the supplied covariance matrix is added as
    regularization term to penalize the loss of the fitted covariance matrix.

    # TODO: add formula.
        source: Modeling Annotation Uncertainty with Gaussian Heatmaps in Landmark Localization

    Args:
        alpha (float, optional): Weight of the regularization term. Defaults to 5.0.
        distance (str, optional): Distance function to use for the loss calculation. Defaults to
            'l2'. Possible values are 'l2', 'l1', 'smooth-l1',
            'bce-with-logits' and 'bce'.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
        **kwargs: Additional keyword arguments for the distance function.
    """

    def __init__(
        self, alpha: float = 5.0, distance: str = "l2", reduction: str = "mean", **kwargs
    ) -> None:
        super().__init__()
        self.alpha = alpha
        self.distance = distance
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        if self.distance == "l2":
            self.dist_func = lambda x, y: F.mse_loss(x, y, reduction="sum")
        elif self.distance == "l1":
            self.dist_func = lambda x, y: F.l1_loss(x, y, reduction="sum")
        elif self.distance == "smooth-l1":
            self.dist_func = lambda x, y: F.smooth_l1_loss(x, y, reduction="sum", **kwargs)
        elif self.distance == "bce-with-logits":
            self.dist_func = lambda x, y: F.binary_cross_entropy_with_logits(x, y, reduction="sum")
        elif self.distance == "bce":
            self.dist_func = lambda x, y: F.binary_cross_entropy(x, y, reduction="sum")
        else:
            raise ValueError(f"Invalid distance function: {self.distance}")

    def forward(
        self, heatmap: torch.Tensor, cov: torch.Tensor, heatmap_target: torch.Tensor
    ) -> torch.Tensor:
        """
        Args:
            heatmap (torch.Tensor): Predicted heatmap.
            cov (torch.Tensor): Covariance matrix of the fitted heatmap.
            heatmap_target (torch.Tensor): Target heatmap.

        Returns:
            torch.Tensor: Loss value.
        """
        det_cov = torch.det(cov)
        assert torch.all(det_cov > 0), f"Determinant of covariance matrix is negative: {det_cov}"
        loss = self.dist_func(heatmap, heatmap_target) / heatmap.shape[0] + self.alpha * torch.sum(
            torch.sqrt(det_cov)
        )
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        return loss


class GaussianHeatmapL2Loss(nn.Module):
    """
    Loss function for Gaussian heatmap regression.
    source: http://arxiv.org/abs/2109.09533"""

    def __init__(self, alpha=5, reduction="mean", spatial_dims=2):
        super(GaussianHeatmapL2Loss, self).__init__()
        self.alpha = alpha
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        self.spatial_dims = spatial_dims
        if spatial_dims not in [2, 3]:
            raise ValueError("spatial_dims must be 2 or 3")

    def forward(self, heatmap, sigmas, heatmap_target):
        if self.spatial_dims == 2:
            loss = F.mse_loss(heatmap, heatmap_target, reduction="sum") / heatmap.shape[
                0
            ] + self.alpha * torch.sum((sigmas[:, 0] * sigmas[:, 1]))
        else:
            loss = F.mse_loss(heatmap, heatmap_target, reduction="sum") / heatmap.shape[
                0
            ] + self.alpha * torch.sum((sigmas[..., 0] * sigmas[..., 1]) * sigmas[..., 2])
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        return loss


class EuclideanDistanceVarianceReg(nn.Module):
    """
    Euclidean distance loss with variance regularization. The regularization term is defined as the
    squared difference between the fitted/predicted variance and a predefined target variance, as
    proposed by Nibali et al. (2018). The authors point out that this regularization term does not
    directly constrain the specif shape of the learned heatmaps.
        source: Numerical Coordinate Regression with Convolutional Neural Networks - Nibali et al.
            (2018)

    Args:
        alpha (float, optional): Weight of the regularization term. Defaults to 1.0.
        var_t (float, optional): Target variance. Defaults to 1.0.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
        eps (float, optional): Epsilon value to avoid division by zero. Defaults to 1e-6.
    """

    def __init__(
        self,
        alpha: float = 1.0,
        var_t: float = 1.0,
        reduction: str = "mean",
        eps: float = 1e-6,
        spatial_dims=2,
    ) -> None:
        super().__init__()
        self.alpha = alpha
        self.var_t = var_t
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        self.eps = eps
        self.spatial_dims = spatial_dims
        if spatial_dims not in [2, 3]:
            raise ValueError("spatial_dims must be 2 or 3")

    def forward(self, pred: torch.Tensor, cov: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            pred (torch.Tensor): Predicted coordinates.
            cov (torch.Tensor): Related covariance matrix of the predicted coordinates.
            target (torch.Tensor): Target coordinates.
        """
        if self.spatial_dims == 2:
            reg = (cov[..., 0, 0] - self.var_t) ** 2 + (cov[..., 1, 1] - self.var_t) ** 2
        else:
            reg = (
                (cov[..., 0, 0] - self.var_t) ** 2
                + (cov[..., 1, 1] - self.var_t) ** 2
                + (cov[..., 2, 2] - self.var_t) ** 2
            )
        loss = _euclidean_distance(pred, target) + self.alpha * reg
        if self.reduction == "mean":
            return torch.mean(loss)
        if self.reduction == "sum":
            return torch.sum(loss)
        return loss


class MultivariateGaussianNLLLoss(nn.Module):
    """
    Negative log-likelihood loss for multivariate Gaussian distributions. The loss function is
    defined as the negative log-likelihood of the predicted coordinates given the predicted
    covariance matrix. The loss function is defined as:
        :math:`NLL = 0.5 * (log(det(Cov)) + (x - mu)^T * Cov^{-1} * (x - mu))`

        As proposed in: "UGLLI Face Alignment: Estimating Uncertainty with Gaussian Log-Likelihood
            Loss" - Kumar et al. (2019)

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
        eps (float, optional): Epsilon value to avoid division by zero. Defaults to 1e-6.
    """

    def __init__(self, reduction: str = "mean", eps: float = 1e-6, spatial_dims: int = 2):
        super().__init__()
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        self.eps = eps
        self.spatial_dims = spatial_dims

    def forward(self, pred: torch.Tensor, cov: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            pred (torch.Tensor): Predicted coordinates.
            cov (torch.Tensor): Related covariance matrix of the predicted coordinates.
            target (torch.Tensor): Target coordinates.
        """
        loss = 0.5 * (
            torch.log(torch.det(cov) + self.eps)
            + (
                (target - pred).unsqueeze(-2)
                @ torch.linalg.inv(cov)
                @ (target - pred).unsqueeze(-2).transpose(-2, -1)
            )
            .squeeze(-1)
            .squeeze(-1)
        )
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        return loss


class WingLoss(nn.Module):
    r"""
    Wing loss, proposed for facial landmark detection by Feng et al. (2018), is a piece-wise loss
    function that focusses more attention on small and medium range erros compared to L2, L1, and
    smooth L1. It has large gradient when the error is small and a constant gradient when the error
    is large.
    The loss function is defined as:

    .. math::
        Wing(x, y) =
        \begin{cases}
            \omega * log(1 + \frac{|x - y|}{\epsilon}) & \text{if } |x - y| < \omega \\
            |x - y| - C & \text{otherwise}
        \end{cases}

    where :math:`C = \omega - \omega * log(1 + \frac{\omega}{\epsilon})`
        source: "Wing Loss for Robust Facial Landmark Localisation With Convolutional Neural
            Networks" - Feng et al. (2018)

    Args:
        omega (float, optional): Wing loss parameter. Defaults to 5.0.
        epsilon (float, optional): Wing loss parameter. Defaults to 0.5.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
    """

    def __init__(
        self, omega: float = 5.0, epsilon: float = 0.5, reduction: str = "mean", spatial_dims=2
    ):
        super().__init__()
        self.omega = omega
        self.epsilon = epsilon
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        self.c = self.omega * (1 - np.log(1 + self.omega / self.epsilon))
        self.spatial_dims = spatial_dims
        if self.spatial_dims == 2:
            raise ValueError("Only 2D heatmaps are supported.")

    def forward(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            pred (torch.Tensor): Predicted coordinates.
            target (torch.Tensor): Target coordinates.
        """
        diff_abs = torch.abs(pred - target)
        loss = torch.where(
            diff_abs < self.omega,
            self.omega * torch.log(1 + diff_abs / self.epsilon),
            diff_abs - self.c,
        )
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        return loss


class AdaptiveWingLoss(nn.Module):
    r"""
    Adaptive wing loss is a loss function that behaves like a smoothed Wing loss when the target is
    close to 1 and like the MSE loss when the target is close to 0.

    The loss function is defined as:

    .. math::
        AWing(x, y) =
        \begin{cases}
            \omega \log(1 + |\frac{x-y}{\epsilon}|^{\alpha-y} &
                \text{if } |x - y| < \theta \\
            A|x - y| - C & \text{otherwise}
        \end{cases}})
    where :math:`A = \omega (1/(1+(\theta/\epsilon)^{\alpha-y}))(\alpha - y)
                    ((\theta/\epsilon)^(\alpha-y-1))(1/\epsilon)` and
                :math:`C = (\theta * A - \omega * log(1 + (\theta/\epsilon)^{\alpha-y})))`
        source: "Adaptive Wing Loss for Robust Face Alignment via Heatmap Regression" - Wang et al.
            (2019)

    Args:
        omega (float, optional): Adaptive wing loss parameter. Defaults to 5.0.
        epsilon (float, optional): Adaptive wing loss parameter. Defaults to 0.5.
        alpha (float, optional): Adaptive wing loss parameter. Defaults to 2.1.
        theta (float, optional): Adaptive wing loss parameter. Defaults to 0.5.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
    """

    def __init__(
        self,
        omega: float = 5,
        epsilon: float = 0.5,
        alpha: float = 2.1,
        theta: float = 0.5,
        reduction: str = "mean",
        spatial_dims=2,
    ) -> None:
        super().__init__()
        self.omega = omega
        self.epsilon = epsilon
        self.alpha = alpha
        self.theta = theta
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        self.spatial_dims = spatial_dims
        if self.spatial_dims != 2:
            raise ValueError("Only 2D heatmaps are supported.")

    def forward(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            pred (torch.Tensor): Predicted coordinates.
            target (torch.Tensor): Target coordinates.
        """
        diff_abs = torch.abs(pred - target)
        a = (
            self.omega
            * (1 / (1 + torch.pow(self.theta / self.epsilon, self.alpha - target)))
            * (self.alpha - target)
            * torch.pow(self.theta / self.epsilon, self.alpha - target - 1)
            * (1 / self.epsilon)
        )
        c = self.theta * a - self.omega * torch.log(
            1 + torch.pow(self.theta / self.epsilon, self.alpha - target)
        )
        loss = torch.where(
            diff_abs < self.theta,
            self.omega * torch.log(1 + torch.pow(diff_abs / self.epsilon, self.alpha - target)),
            a * diff_abs - c,
        )
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        return loss


class StarLoss(nn.Module):
    """
    Self-adapTive Ambiguity Reduction (STAR) loss. Star loss takes into account the ambuigity
    (uncertainty) of the intermediate heatmap predictions by using the covariance matrix of the
    heatmap predictions, and extracting the eigenvectors and eigenvalues of the covariance matrix.

        source: "STAR Loss: Reducing Semantic Ambiguity in Facial Landmark Detection" - Zhou et al.
            (2023)

    Args:
        omega (float, optional): Weight of the regularization term. Defaults to 1.0.
        distance (str, optional): Distance function to use for the loss calculation. Defaults to
            'l2'. Possible values are 'l2', 'l1', 'smooth-l1'.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
        **kwargs: Additional keyword arguments for the distance function.
    """

    def __init__(
        self,
        omega: float = 1.0,
        distance: str = "l2",
        epsilon: float = 1e-5,
        reduction: str = "mean",
        spatial_dims=2,
        **kwargs,
    ) -> None:
        super().__init__()
        self.omega = omega
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")
        self.distance = distance
        self.spatial_dims = spatial_dims
        if self.distance == "l2":
            self.dist_func = lambda x, y: F.mse_loss(x, y, reduction="sum")
        elif self.distance == "l1":
            self.dist_func = lambda x, y: F.l1_loss(x, y, reduction="sum")
        elif self.distance == "smooth-l1":
            self.dist_func = lambda x, y: F.smooth_l1_loss(x, y, reduction="sum", **kwargs)
        else:
            raise ValueError(f"Invalid distance function: {self.distance}")
        self.epsilon = epsilon

    def forward(self, pred: torch.Tensor, cov: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            pred (torch.Tensor): Predicted coordinates.
            cov (torch.Tensor): Related covariance matrix of the predicted coordinates.
            target (torch.Tensor): Target coordinates.
        """
        eig_values, eig_vecs = torch.linalg.eig(cov)
        eig_values = eig_values.float()
        eig_vecs = eig_vecs.float()
        if self.spatial_dims == 2:
            loss = (1 / torch.sqrt(eig_values[..., 0]) + self.epsilon) * self.dist_func(
                eig_vecs[..., 0].unsqueeze(-2) @ pred.unsqueeze(-1),
                eig_vecs[..., 0].unsqueeze(-2) @ target.unsqueeze(-1),
            ) + (1 / torch.sqrt(eig_values[..., 1]) + self.epsilon) * self.dist_func(
                eig_vecs[..., 1].unsqueeze(-2) @ pred.unsqueeze(-1),
                eig_vecs[..., 1].unsqueeze(-2) @ target.unsqueeze(-1),
            )
        else:
            loss = (
                (1 / torch.sqrt(eig_values[..., 0]) + self.epsilon)
                * self.dist_func(
                    eig_vecs[..., 0].unsqueeze(-2) @ pred.unsqueeze(-1),
                    eig_vecs[..., 0].unsqueeze(-2) @ target.unsqueeze(-1),
                )
                + (1 / torch.sqrt(eig_values[..., 1]) + self.epsilon)
                * self.dist_func(
                    eig_vecs[..., 1].unsqueeze(-2) @ pred.unsqueeze(-1),
                    eig_vecs[..., 1].unsqueeze(-2) @ target.unsqueeze(-1),
                )
                + (1 / torch.sqrt(eig_values[..., 2]) + self.epsilon)
                * self.dist_func(
                    eig_vecs[..., 2].unsqueeze(-2) @ pred.unsqueeze(-1),
                    eig_vecs[..., 2].unsqueeze(-2) @ target.unsqueeze(-1),
                )
            )
        loss += self.omega / 2 * torch.sum(eig_values, dim=-1)
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        return loss


class NLLLoss(nn.Module):
    """Negative log-likelihood loss for 2D/3D heatmaps. Assumes that the input is a probability
    distribution and calculates the negative log-likelihood of the predicted heatmap given the
    target heatmap.

    Args:
        spatial_dims (int, optional): Spatial dimension of the heatmaps. Defaults to 2.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
    """

    def __init__(self, spatial_dims: int = 2, apply_softmax: bool = True, reduction: str = "mean"):
        super().__init__()
        self.spatial_dims = spatial_dims
        self.apply_softmax = apply_softmax
        if self.apply_softmax:
            self.log_softmax = LogSoftmaxND(spatial_dims)
        if spatial_dims not in [2, 3]:
            raise ValueError("spatial_dims must be 2 or 3")
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")

    def forward(self, output, target):
        if self.apply_softmax:
            output = self.log_softmax(output)
        else:
            output = torch.log(output)
        nll = -target * output
        if self.spatial_dims == 2:
            dim = (2, 3)
        else:
            dim = (2, 3, 4)
        if self.reduction == "mean":
            return torch.mean(torch.sum(nll, dim=dim))
        if self.reduction == "sum":
            return torch.sum(torch.sum(nll, dim=dim))
        return torch.sum(nll, dim=dim)


class StackedLoss(nn.Module):
    """Stacked loss function. Applies a specified loss function to each list of predictions. This
    loss function is used to calculate the loss for each heatmap in the stacked heatmap regression,
    suchs as stacked hourglass and U-Net networks.

    Args:
        loss_fn (nn.Module): Loss function to use.
        reduction (str, optional): Specifies the reduction to apply to the output. Defaults to
            'mean'.
    """

    def __init__(self, loss_fn: nn.Module, reduction: str = "mean", **kwargs) -> None:
        super().__init__()
        self.loss_fn = loss_fn(reduction=reduction, **kwargs)
        self.reduction = reduction
        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(f"Invalid reduction: {reduction}")

    def forward(self, preds: list[torch.Tensor], target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            preds(list[torch.Tensor]): List of predicted heatmaps or coordinates.
            target (torch.Tensor): Target heatmap.
        """
        losses = []
        for pred in preds:
            losses.append(self.loss_fn(pred, target))
        if self.reduction == "mean":
            return torch.stack(losses).mean()
        if self.reduction == "sum":
            return torch.stack(losses).sum()
        return torch.stack(losses)


def _euclidean_distance(x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
    """Computes the euclidean distance between two tensors."""
    return torch.sqrt(torch.sum((x1 - x2) ** 2, dim=-1))



"""Test data module."""

import os

import cv2
import numpy as np
import pytest
import torch
from monai.transforms import (
    Compose,
    RandAdjustContrastd,
    RandAffined,
    RandGaussianNoised,
    RandHistogramShiftd,
    RandScaleIntensityd,
    RandStdShiftIntensityd,
)

from src.landmarker.data import LandmarkDataset
from src.landmarker.data.landmark_dataset import HeatmapDataset, MaskDataset

NSIM = 10  # number of simulated images


@pytest.fixture(scope="session", autouse=True)
def setup_data():
    """Create data for tests."""
    # Check if data folder exists and create it if not
    if not os.path.exists("tests/data"):
        os.makedirs("tests/data")
    all_landmarks = []
    all_img_uint8 = []
    for i in range(NSIM):
        img_uint8 = (np.random.rand(64, 64) * 256).astype(np.uint8)
        img_uint16 = (np.random.random((64, 64)) * 65535).astype(np.uint16)
        img_color = (np.random.rand(64, 64, 3) * 256).astype(np.uint8)
        landmarks = np.random.rand(8, 2)
        landmarks = landmarks * 64
        cv2.imwrite(f"tests/data/{i}_img_uint8.png", img_uint8)
        cv2.imwrite(
            f"tests/data/{i}_img_uint16.png",
            img_uint16,
        )
        cv2.imwrite(f"tests/data/{i}_img_color.png", img_color)
        all_landmarks.append(landmarks)
        all_img_uint8.append(img_uint8.reshape((1, 64, 64)))

    all_landmarks = np.stack(all_landmarks)

    np.save("tests/data/landmarks.npy", all_landmarks)
    np.save("all_img_uint8.npy", all_img_uint8)

    # define global variables
    pytest.landmarks = all_landmarks
    pytest.all_img_uint8 = all_img_uint8

    def clear_data():
        """Remove data for tests."""
        for i in range(NSIM):
            os.remove(f"tests/data/{i}_img_uint8.png")
            os.remove(f"tests/data/{i}_img_uint16.png")
            os.remove(f"tests/data/{i}_img_color.png")
        os.remove("tests/data/landmarks.npy")
        os.remove("all_img_uint8.npy")

    yield
    clear_data()


@pytest.fixture(scope="session", autouse=True)
def setup_data_3d():
    """Create data 3d for tests."""
    # Check if data folder exists and create it if not
    if not os.path.exists("tests/data"):
        os.makedirs("tests/data")
    all_landmarks = []
    all_img_uint8 = []
    for i in range(NSIM):
        img_uint8 = (np.random.rand(64, 64, 64) * 256).astype(np.uint8)
        img_uint16 = (np.random.random((64, 64, 64)) * 65535).astype(np.uint16)
        np.save(f"tests/data/{i}_img_3d_uint8.npy", img_uint8)
        np.save(f"tests/data/{i}_img_3d_uint16.npy", img_uint16)
        landmarks = np.random.rand(8, 3)
        landmarks = landmarks * 64
        all_landmarks.append(landmarks)

    all_landmarks = np.stack(all_landmarks)

    np.save("tests/data/landmarks_3d.npy", all_landmarks)

    # define global variables
    pytest.landmarks_3d = all_landmarks
    pytest.all_img_uint8_3d = all_img_uint8

    def clear_data():
        """Remove data for tests."""
        for i in range(NSIM):
            os.remove(f"tests/data/{i}_img_3d_uint8.npy")
            os.remove(f"tests/data/{i}_img_3d_uint16.npy")
        os.remove("tests/data/landmarks_3d.npy")

    yield
    clear_data()


def test_landmark_dataset_stored_imgs():
    """Test LandmarkDatase with store_imgs=True."""
    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        class_names=[f"class_{i}" for i in range(8)],
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint16.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        dim_img=(32, 20),
        pixel_spacing=(0.1, 0.1),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint16.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        dim_img=(32, 32),
        resize_pad=False,
        pixel_spacing=torch.Tensor([[0.1, 0.1]]).repeat(NSIM, 1),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 32])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 32])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=pytest.all_img_uint8,
        landmarks=pytest.landmarks,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    all_img_uint8_tensor_list = [torch.Tensor(img) for img in pytest.all_img_uint8]
    dataset = LandmarkDataset(
        imgs=all_img_uint8_tensor_list,
        landmarks=pytest.landmarks,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    all_img_uint8_tensor_list = [torch.Tensor(img) for img in pytest.all_img_uint8]
    dataset = LandmarkDataset(
        imgs=torch.stack(all_img_uint8_tensor_list),
        landmarks=torch.Tensor(pytest.landmarks),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=pytest.all_img_uint8,
        landmarks=pytest.landmarks,
        dim_img=(32, 20),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=pytest.all_img_uint8, landmarks=pytest.landmarks, dim_img=(32, 32), resize_pad=False
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 32])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 32])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark


def test_landmark_dataset_not_store_img():
    """Test LandmarkDataset with images and store_imgs=False."""
    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        store_imgs=False,
        dim_img=(32, 20),
        resize_pad=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        store_imgs=False,
        dim_img=(30, 15),
        resize_pad=True,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 30, 15])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 30, 15])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark
    assert torch.allclose(dataset[0]["padding"], torch.Tensor((32, 0)))  # padding


def test_landmark_dataset():
    """Test LandmarkDataset with images."""
    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_color.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([3, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_color.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        dim_img=(32, 20),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([3, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_color.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([3, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_color.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        dim_img=(32, 20),
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([3, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark

    all_img_uint8_tensor_list = [torch.Tensor(img) for img in pytest.all_img_uint8]
    dataset = LandmarkDataset(
        imgs=all_img_uint8_tensor_list,
        landmarks=pytest.landmarks,
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    all_img_uint8_tensor_list = [torch.Tensor(img) for img in pytest.all_img_uint8]
    dataset = LandmarkDataset(
        imgs=torch.stack(all_img_uint8_tensor_list),
        landmarks=torch.Tensor(pytest.landmarks),
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=pytest.all_img_uint8,
        landmarks=pytest.landmarks,
        dim_img=(32, 20),
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=pytest.all_img_uint8,
        landmarks=pytest.landmarks,
        dim_img=(32, 32),
        resize_pad=False,
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 32])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 32])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark


def test_landmark_dataset_transforms():
    """Test LandmarkDataset with transforms."""
    fn_keys = ("image",)
    spatial_transformd = [
        RandAffined(
            fn_keys,
            prob=0.75,
            rotate_range=(-np.pi / 12, np.pi / 12),
            translate_range=(-100, 100),
            scale_range=(-0.25, 0.25),
        )
    ]

    composed_transformd = Compose(
        spatial_transformd
        +
        # Add gaussian noise
        [
            RandGaussianNoised(("image",), prob=0.1, mean=0, std=0.005),
            # Add random intensity shift
            RandStdShiftIntensityd(("image",), prob=0.1, factors=0.1),
            # Add random intensity scaling
            RandScaleIntensityd(("image",), factors=0.25, prob=0.1),
            RandAdjustContrastd(("image",), prob=0.1, gamma=(0.5, 2)),  # Randomly adjust contrast
            # Randomly shift histogram
            RandHistogramShiftd(("image",), prob=0.1),
        ]
    )
    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        transform=composed_transformd,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        transform=composed_transformd,
        dim_img=(32, 20),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        transform=composed_transformd,
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = LandmarkDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        transform=composed_transformd,
        dim_img=(32, 20),
        store_imgs=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark


def test_mask_landmark_dataset():
    """Test MaskDataset."""
    dataset = MaskDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 64, 64])  # mask
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 64, 64]
    )  # mask
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    for i, batch in enumerate(dataset):
        cv2.imwrite(f"tests/data/{i}_mask.png", batch["mask"].sum(dim=0).numpy() * 255)
    dataset = MaskDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        mask_paths=[f"tests/data/{i}_mask.png" for i in range(NSIM)],
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["mask"].shape == torch.Size([1, 64, 64])  # mask
    assert dataset[0]["landmark"].shape == torch.Size([1, 8, 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size([1, 64, 64])  # mask
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size([1, 8, 2])  # landmark

    dataset = MaskDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        dim_img=(32, 20),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 32, 20])  # mask
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 32, 20]
    )  # mask
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark

    dataset = MaskDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        mask_paths=[f"tests/data/{i}_mask.png" for i in range(NSIM)],
        dim_img=(32, 20),
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["mask"].shape == torch.Size([1, 32, 20])  # mask
    assert dataset[0]["landmark"].shape == torch.Size([1, 8, 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size([1, 32, 20])  # mask
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size([1, 8, 2])  # landmark

    dataset = MaskDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        mask_paths=[f"tests/data/{i}_mask.png" for i in range(NSIM)],
        dim_img=(32, 20),
        resize_pad=False,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[0]["mask"].shape == torch.Size([1, 32, 20])  # mask
    assert dataset[0]["landmark"].shape == torch.Size([1, 8, 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 32, 20])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size([1, 32, 20])  # mask
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size([1, 8, 2])  # landmark

    # remove masks
    for i in range(NSIM):
        os.remove(f"tests/data/{i}_mask.png")


def test_mask_landmark_transforms():
    """Test MaskDataset with transforms."""
    fn_keys = ("image",)
    spatial_transformd = [
        RandAffined(
            fn_keys,
            prob=0.75,
            rotate_range=(-np.pi / 12, np.pi / 12),
            translate_range=(-100, 100),
            scale_range=(-0.25, 0.25),
        )
    ]

    composed_transformd = Compose(
        spatial_transformd
        +
        # Add gaussian noise
        [
            RandGaussianNoised(("image",), prob=0.1, mean=0, std=0.005),
            # Add random intensity shift
            RandStdShiftIntensityd(("image",), prob=0.1, factors=0.1),
            # Add random intensity scaling
            RandScaleIntensityd(("image",), factors=0.25, prob=0.1),
            RandAdjustContrastd(("image",), prob=0.1, gamma=(0.5, 2)),  # Randomly adjust contrast
            # Randomly shift histogram
            RandHistogramShiftd(("image",), prob=0.1),
        ]
    )

    dataset = MaskDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        transform=composed_transformd,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 64, 64])  # mask
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 64, 64]
    )  # mask
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark


def test_heatmap_dataset():
    """Test HeatmapDataset."""
    dataset = HeatmapDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        heatmap_fun="gaussian",
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 64, 64])  # heatmap
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 64, 64]
    )  # heatmap
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark
    dataset = HeatmapDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        heatmap_fun="laplacian",
        dim_img=(20, 32),
        batch_size=0,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 20, 32])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 20, 32])  # heatmap
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 20, 32])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 20, 32]
    )  # heatmap
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark


def test_heatmap_dataset_transforms():
    """Test HeatmapDataset with transforms."""
    fn_keys = ("image", "mask")
    spatial_transformd = [
        RandAffined(
            fn_keys,
            prob=0.75,
            rotate_range=(-np.pi / 12, np.pi / 12),
            translate_range=(-100, 100),
            scale_range=(-0.25, 0.25),
        )
    ]

    composed_transformd = Compose(
        spatial_transformd
        +
        # Add gaussian noise
        [
            RandGaussianNoised(("image",), prob=0.1, mean=0, std=0.005),
            # Add random intensity shift
            RandStdShiftIntensityd(("image",), prob=0.1, factors=0.1),
            # Add random intensity scaling
            RandScaleIntensityd(("image",), factors=0.25, prob=0.1),
            RandAdjustContrastd(("image",), prob=0.1, gamma=(0.5, 2)),  # Randomly adjust contrast
            # Randomly shift histogram
            RandHistogramShiftd(("image",), prob=0.1),
        ]
    )

    dataset = HeatmapDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        heatmap_fun="gaussian",
        transform=composed_transformd,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 64, 64])  # heatmap
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 64, 64])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 64, 64]
    )  # heatmap
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark
    dataset = HeatmapDataset(
        imgs=[f"tests/data/{i}_img_uint8.png" for i in range(NSIM)],
        landmarks=pytest.landmarks,
        heatmap_fun="laplacian",
        dim_img=(20, 32),
        batch_size=0,
        transform=composed_transformd,
    )
    assert len(dataset) == len(pytest.landmarks)
    assert dataset[0]["image"].shape == torch.Size([1, 20, 32])  # image
    assert dataset[0]["mask"].shape == torch.Size([pytest.landmarks.shape[1], 20, 32])  # heatmap
    assert dataset[0]["landmark"].shape == torch.Size([pytest.landmarks.shape[1], 2])  # landmark
    assert dataset[len(pytest.landmarks) - 1]["image"].shape == torch.Size([1, 20, 32])  # image
    assert dataset[len(pytest.landmarks) - 1]["mask"].shape == torch.Size(
        [pytest.landmarks.shape[1], 20, 32]
    )  # heatmap
    assert dataset[len(pytest.landmarks) - 1]["landmark"].shape == torch.Size(
        [pytest.landmarks.shape[1], 2]
    )  # landmark



"""
Heatmap module.
"""

from .decoder import (
    coord_argmax,
    coord_cov_from_gaussian_ls,
    coord_local_soft_argmax,
    coord_weighted_spatial_mean,
    coord_weighted_spatial_mean_cov,
    cov_from_gaussian_ls,
    heatmap_to_coord,
    heatmap_to_coord_cov,
    heatmap_to_coord_enlarge,
    heatmap_to_multiple_coord,
    non_maximum_surpression,
    weighted_sample_cov,
    windowed_weigthed_sample_cov,
)
from .generator import GaussianHeatmapGenerator, LaplacianHeatmapGenerator

__all__ = [
    "LaplacianHeatmapGenerator",
    "GaussianHeatmapGenerator",
    "coord_argmax",
    "coord_local_soft_argmax",
    "coord_weighted_spatial_mean",
    "heatmap_to_coord",
    "heatmap_to_coord_enlarge",
    "coord_soft_argmax_cov",
    "coord_weighted_spatial_mean_cov",
    "heatmap_to_coord_cov",
    "coord_cov_from_gaussian_ls",
    "cov_from_gaussian_ls",
    "weighted_sample_cov",
    "windowed_weigthed_sample_cov",
    "heatmap_to_multiple_coord",
    "non_maximum_surpression",
    "non_maximum_surpression_local_soft_argmax",
    "softmax_2d",
]



"""Test for the loss module."""

import torch
import torch.nn as nn

from src.landmarker.heatmap.generator import GaussianHeatmapGenerator
from src.landmarker.losses.losses import (
    EuclideanDistanceVarianceReg,
    GeneralizedNormalHeatmapLoss,
    MultivariateGaussianNLLLoss,
    NLLLoss,
    StackedLoss,
    StarLoss,
)


def test_generalized_normal_heatmap_loss():
    """Test the GeneralizedNormalHeatmapLoss class."""
    # create an instance of the loss function

    for distance in ["l1", "l2", "smooth-l1", "bce-with-logits", "bce"]:
        if "l1" == distance:
            reduction = "mean"
        else:
            reduction = "sum"
        loss_fn = GeneralizedNormalHeatmapLoss(alpha=5, distance=distance, reduction=reduction)

        # create some dummy input tensors
        if distance in "bce":
            pred = torch.rand(2, 3, 64, 64)
            target = torch.rand(2, 3, 64, 64)
        elif distance in ["bce-with-logits"]:
            pred = torch.randn(2, 3, 64, 64)
            target = torch.rand(2, 3, 64, 64)
        else:
            pred = torch.rand(2, 3, 64, 64) * 64
            target = torch.rand(2, 3, 64, 64) * 64
        sigmas = torch.rand(3, 2) * 5
        rotation = torch.rand(3) * 2 * torch.pi
        heatmap_generator = GaussianHeatmapGenerator(
            3, sigmas=sigmas, heatmap_size=(64, 64), rotation=rotation
        )
        covs = heatmap_generator.get_covariance_matrix()

        # calculate the loss
        loss = loss_fn(pred, covs, target)

        # check that the output has the correct shape
        assert loss.shape == torch.Size([])

        # check that the output is a scalar tensor
        assert loss.dim() == 0

        # check that the output is non-negative
        assert loss >= 0
    try:
        loss_fn = GeneralizedNormalHeatmapLoss(alpha=5, distance="invalid", reduction="mean")
        assert False
    except ValueError:
        assert True


def test_generalized_normal_heatmap_loss_3d():
    """Test the GeneralizedNormalHeatmapLoss class. For 3D inputs."""
    # create an instance of the loss function

    for distance in ["l1", "l2", "smooth-l1", "bce-with-logits", "bce"]:
        if "l1" == distance:
            reduction = "mean"
        else:
            reduction = "sum"
        loss_fn = GeneralizedNormalHeatmapLoss(alpha=5, distance=distance, reduction=reduction)

        # create some dummy input tensors
        if distance in "bce":
            pred = torch.rand(2, 3, 64, 64, 64)
            target = torch.rand(2, 3, 64, 64, 64)
        elif distance in ["bce-with-logits"]:
            pred = torch.randn(2, 3, 64, 64, 64)
            target = torch.rand(2, 3, 64, 64, 64)
        else:
            pred = torch.rand(2, 3, 64, 64, 64) * 64
            target = torch.rand(2, 3, 64, 64, 64) * 64
        sigmas = torch.rand(3, 3) * 5
        rotation = torch.rand(3, 3) * 2 * torch.pi
        heatmap_generator = GaussianHeatmapGenerator(
            3, sigmas=sigmas, heatmap_size=(64, 64, 64), rotation=rotation
        )
        covs = heatmap_generator.get_covariance_matrix()

        # calculate the loss
        loss = loss_fn(pred, covs, target)

        # check that the output has the correct shape
        assert loss.shape == torch.Size([])

        # check that the output is a scalar tensor
        assert loss.dim() == 0

        # check that the output is non-negative
        assert loss >= 0
    try:
        loss_fn = GeneralizedNormalHeatmapLoss(alpha=5, distance="invalid", reduction="mean")
        assert False
    except ValueError:
        assert True


def test_multivariate_gaussian_nll_loss():
    """Test the MultivariateGaussianNLLLoss class."""
    reduction = "mean"
    pred = torch.rand(1, 3, 2) * 64
    target = torch.rand(1, 3, 2) * 64
    cov_matrix = torch.tensor([[1.0, 0.0], [0.0, 1.0]]).view((1, 1, 2, 2)).repeat(1, 3, 1, 1)

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    pred = torch.rand((4, 5, 2)) * 64
    target = torch.rand((4, 5, 2)) * 64
    cov_matrix = torch.tensor([[1.0, 0.0], [0.0, 1.0]]).view((1, 1, 2, 2)).repeat(4, 5, 1, 1)

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "sum"

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "none"

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction)
    expected_output_shape = torch.Size([4, 5])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape


def test_multivariate_gaussian_nll_loss_3d():
    """Test the MultivariateGaussianNLLLoss class. For 3D inputs."""
    reduction = "mean"
    pred = torch.rand(1, 3, 3) * 64
    target = torch.rand(1, 3, 3) * 64
    cov_matrix = (
        torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        .view((1, 1, 3, 3))
        .repeat(1, 3, 1, 1)
    )

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    pred = torch.rand((4, 5, 3)) * 64
    target = torch.rand((4, 5, 3)) * 64
    cov_matrix = (
        torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        .view((1, 1, 3, 3))
        .repeat(4, 5, 1, 1)
    )

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "sum"

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "none"

    loss_fn = MultivariateGaussianNLLLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([4, 5])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape


def test_euclidean_distance_variance_reg():
    """Test the EuclideanDistanceVarianceReg class."""
    reduction = "mean"
    sigma_t = 3
    # pred = torch.rand(1, 3, 2) * 64
    # target = torch.rand(1, 3, 2) * 64
    pred = torch.ones((1, 3, 2)) * (64 // 2)
    target = torch.ones((1, 3, 2)) * (64 // 2 - 5)

    heatmap_generator = GaussianHeatmapGenerator(3, sigmas=sigma_t, heatmap_size=(64, 64))
    cov_t = heatmap_generator.get_covariance_matrix()

    loss_fn = EuclideanDistanceVarianceReg(reduction=reduction, var_t=sigma_t**2)
    expected_output_shape = torch.Size([])

    loss = loss_fn(target, cov_t, target)

    assert loss.shape == expected_output_shape
    assert loss == 0

    loss = loss_fn(pred, cov_t, target)
    assert loss > 0

    reduction = "sum"
    sigma_t = 3
    # pred = torch.rand(1, 3, 2) * 64
    # target = torch.rand(1, 3, 2) * 64
    pred = torch.ones((1, 3, 2)) * (64 // 2)
    target = torch.ones((1, 3, 2)) * (64 // 2 - 5)

    heatmap_generator = GaussianHeatmapGenerator(3, sigmas=sigma_t, heatmap_size=(64, 64))
    cov_t = heatmap_generator.get_covariance_matrix()

    loss_fn = EuclideanDistanceVarianceReg(reduction=reduction, var_t=sigma_t**2)
    expected_output_shape = torch.Size([])

    loss = loss_fn(target, cov_t, target)

    assert loss.shape == expected_output_shape
    assert loss == 0

    loss = loss_fn(pred, cov_t, target)
    assert loss > 0

    reduction = "none"
    sigma_t = 3
    # pred = torch.rand(1, 3, 2) * 64
    # target = torch.rand(1, 3, 2) * 64
    pred = torch.ones((1, 3, 2)) * (64 // 2)
    target = torch.ones((1, 3, 2)) * (64 // 2 - 5)

    heatmap_generator = GaussianHeatmapGenerator(3, sigmas=sigma_t, heatmap_size=(64, 64))
    cov_t = heatmap_generator.get_covariance_matrix()

    loss_fn = EuclideanDistanceVarianceReg(reduction=reduction, var_t=sigma_t**2)
    expected_output_shape = target.shape[:-1]

    loss = loss_fn(target, cov_t, target)

    assert loss.shape == expected_output_shape
    assert torch.allclose(loss, torch.zeros_like(loss))

    loss = loss_fn(pred, cov_t, target)
    assert torch.all(loss > 0)

    try:
        loss_fn = EuclideanDistanceVarianceReg(reduction="invalid", var_t=-1)
        loss = loss_fn(pred, cov_t, target)
        assert False
    except ValueError:
        assert True


def test_euclidean_distance_variance_reg_3d():
    """Test the EuclideanDistanceVarianceReg class. For 3D inputs."""
    reduction = "mean"
    sigma_t = 3

    pred = torch.ones((1, 3, 3)) * (64 // 2)
    target = torch.ones((1, 3, 3)) * (64 // 2 - 5)

    heatmap_generator = GaussianHeatmapGenerator(3, sigmas=sigma_t, heatmap_size=(64, 64, 64))
    cov_t = heatmap_generator.get_covariance_matrix()

    loss_fn = EuclideanDistanceVarianceReg(reduction=reduction, var_t=sigma_t**2, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(target, cov_t, target)

    assert loss.shape == expected_output_shape
    assert loss == 0

    loss = loss_fn(pred, cov_t, target)
    assert loss > 0

    reduction = "sum"
    sigma_t = 3

    pred = torch.ones((1, 3, 3)) * (64 // 2)
    target = torch.ones((1, 3, 3)) * (64 // 2 - 5)

    heatmap_generator = GaussianHeatmapGenerator(3, sigmas=sigma_t, heatmap_size=(64, 64, 64))
    cov_t = heatmap_generator.get_covariance_matrix()

    loss_fn = EuclideanDistanceVarianceReg(reduction=reduction, var_t=sigma_t**2, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(target, cov_t, target)

    assert loss.shape == expected_output_shape
    assert loss == 0

    loss = loss_fn(pred, cov_t, target)
    assert loss > 0

    reduction = "none"
    sigma_t = 3

    pred = torch.ones((1, 3, 3)) * (64 // 2)
    target = torch.ones((1, 3, 3)) * (64 // 2 - 5)

    heatmap_generator = GaussianHeatmapGenerator(3, sigmas=sigma_t, heatmap_size=(64, 64, 64))
    cov_t = heatmap_generator.get_covariance_matrix()

    loss_fn = EuclideanDistanceVarianceReg(reduction=reduction, var_t=sigma_t**2, spatial_dims=3)
    expected_output_shape = target.shape[:-1]

    loss = loss_fn(target, cov_t, target)

    assert loss.shape == expected_output_shape
    assert torch.allclose(loss, torch.zeros_like(loss))

    loss = loss_fn(pred, cov_t, target)
    assert torch.all(loss > 0)

    try:
        loss_fn = EuclideanDistanceVarianceReg(reduction="invalid", var_t=-1, spatial_dims=3)
        loss = loss_fn(pred, cov_t, target)
        assert False
    except ValueError:
        assert True


def test_star_loss():
    """Test the StarLoss class.""" ""
    reduction = "mean"
    pred = torch.rand(1, 3, 2) * 64
    target = torch.rand(1, 3, 2) * 64
    cov_matrix = torch.tensor([[1.0, 0.0], [0.0, 1.0]]).view((1, 1, 2, 2)).repeat(1, 3, 1, 1)

    loss_fn = StarLoss(reduction=reduction)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    pred = torch.rand((4, 5, 2)) * 64
    target = torch.rand((4, 5, 2)) * 64
    cov_matrix = torch.tensor([[1.0, 0.0], [0.0, 1.0]]).view((1, 1, 2, 2)).repeat(4, 5, 1, 1)

    loss_fn = StarLoss(reduction=reduction)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "sum"

    loss_fn = StarLoss(reduction=reduction)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "none"

    loss_fn = StarLoss(reduction=reduction)
    expected_output_shape = torch.Size([4, 5])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    for distance in ["l1", "l2", "smooth-l1"]:
        loss_fn = StarLoss(reduction=reduction, distance=distance)
        expected_output_shape = torch.Size([4, 5])

        loss = loss_fn(pred, cov_matrix, target)

        assert loss.shape == expected_output_shape

    try:
        loss_fn = StarLoss(reduction=reduction, distance="invalid")
        assert False
    except ValueError:
        assert True


def test_star_loss_3d():
    """Test the StarLoss class. For 3D inputs."""
    reduction = "mean"
    pred = torch.rand(1, 3, 3) * 64
    target = torch.rand(1, 3, 3) * 64
    cov_matrix = (
        torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        .view((1, 1, 3, 3))
        .repeat(1, 3, 1, 1)
    )

    loss_fn = StarLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    pred = torch.rand((4, 5, 3)) * 64
    target = torch.rand((4, 5, 3)) * 64
    cov_matrix = (
        torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        .view((1, 1, 3, 3))
        .repeat(4, 5, 1, 1)
    )

    loss_fn = StarLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "sum"

    loss_fn = StarLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    reduction = "none"

    loss_fn = StarLoss(reduction=reduction, spatial_dims=3)
    expected_output_shape = torch.Size([4, 5])

    loss = loss_fn(pred, cov_matrix, target)

    assert loss.shape == expected_output_shape

    for distance in ["l1", "l2", "smooth-l1"]:
        loss_fn = StarLoss(reduction=reduction, distance=distance, spatial_dims=3)
        expected_output_shape = torch.Size([4, 5])

        loss = loss_fn(pred, cov_matrix, target)

        assert loss.shape == expected_output_shape

    try:
        loss_fn = StarLoss(reduction=reduction, distance="invalid")
        assert False
    except ValueError:
        assert True


def test_stacked_loss():
    """Test the StackedLoss class."""
    # create an instance of the stacked loss function with mean reduction
    loss_fn = nn.MSELoss
    stacked_loss_fn = StackedLoss(loss_fn, reduction="mean")

    # create some dummy input tensors

    preds = [torch.rand(2, 4, 64, 64) for _ in range(3)]
    target = torch.rand(2, 4, 64, 64)

    # calculate the loss
    loss = stacked_loss_fn(preds, target)

    # check that the output has the correct shape
    assert loss.shape == torch.Size([])

    # check that the output is a scalar tensor
    assert loss.dim() == 0

    # check that the output is non-negative
    assert loss >= 0

    # create an instance of the stacked loss function with sum reduction
    stacked_loss_fn_sum = StackedLoss(loss_fn, reduction="sum")

    # calculate the loss
    loss_sum = stacked_loss_fn_sum(preds, target)

    # check that the output has the correct shape
    assert loss_sum.shape == torch.Size([])

    # check that the output is a scalar tensor
    assert loss_sum.dim() == 0

    # check that the output is non-negative
    assert loss_sum >= 0

    # check that the sum reduction is equivalent to 2*4*64*64 times the mean reduction
    assert torch.allclose(loss_sum, 2 * 3 * 4 * 64 * 64 * loss)

    # create an instance of the stacked loss function with element-wise reduction
    stacked_loss_fn_elements = StackedLoss(loss_fn, reduction="none")

    # calculate the loss
    loss = stacked_loss_fn_elements(preds, target)

    # check that the output has the correct shape
    assert loss.shape == torch.Size([3, 2, 4, 64, 64])

    # check that the output is non-negative
    assert (loss >= 0).all()


def test_stacked_loss_3d():
    """Test the StackedLoss class. For 3D inputs."""
    # create an instance of the stacked loss function with mean reduction
    loss_fn = nn.MSELoss
    stacked_loss_fn = StackedLoss(loss_fn, reduction="mean")

    # create some dummy input tensors

    preds = [torch.rand(2, 4, 64, 64, 64) for _ in range(3)]
    target = torch.rand(2, 4, 64, 64, 64)

    # calculate the loss
    loss = stacked_loss_fn(preds, target)

    # check that the output has the correct shape
    assert loss.shape == torch.Size([])

    # check that the output is a scalar tensor
    assert loss.dim() == 0

    # check that the output is non-negative
    assert loss >= 0

    # create an instance of the stacked loss function with sum reduction
    stacked_loss_fn_sum = StackedLoss(loss_fn, reduction="sum")

    # calculate the loss
    loss_sum = stacked_loss_fn_sum(preds, target)

    # check that the output has the correct shape
    assert loss_sum.shape == torch.Size([])

    # check that the output is a scalar tensor
    assert loss_sum.dim() == 0

    # check that the output is non-negative
    assert loss_sum >= 0

    # check that the sum reduction is equivalent to 2*4*64*64 times the mean reduction
    assert torch.allclose(loss_sum, 2 * 3 * 4 * 64 * 64 * 64 * loss)

    # create an instance of the stacked loss function with element-wise reduction
    stacked_loss_fn_elements = StackedLoss(loss_fn, reduction="none")

    # calculate the loss
    loss = stacked_loss_fn_elements(preds, target)

    # check that the output has the correct shape
    assert loss.shape == torch.Size([3, 2, 4, 64, 64, 64])

    # check that the output is non-negative
    assert (loss >= 0).all()


def test_NLLLoss_2d():
    """Test the NLLLoss class."""
    reduction = "mean"
    pred = torch.rand(2, 3, 64, 64)
    target = torch.rand(2, 3, 64, 64)

    loss_fn = NLLLoss(reduction=reduction)
    expected_output_shape = torch.Size([])
    loss = loss_fn(pred, target)
    assert loss.shape == expected_output_shape

    pred = torch.rand(2, 3, 64, 64)
    target = torch.rand(2, 3, 64, 64)

    loss_fn = NLLLoss(reduction="sum")
    expected_output_shape = torch.Size([])

    loss = loss_fn(pred, target)
    assert loss.shape == expected_output_shape

    loss_fn = NLLLoss(reduction="none")
    expected_output_shape = torch.Size([2, 3])

    loss = loss_fn(pred, target)
    assert loss.shape == expected_output_shape



"""Tests for the decoder module."""

import numpy as np
import torch

from src.landmarker.heatmap.decoder import (
    coord_argmax,
    coord_cov_from_gaussian_ls,
    coord_local_soft_argmax,
    coord_weighted_spatial_mean,
    cov_from_gaussian_ls,
    heatmap_to_coord,
    heatmap_to_coord_cov,
    heatmap_to_coord_enlarge,
    heatmap_to_multiple_coord,
    weighted_sample_cov,
    windowed_weigthed_sample_cov,
)
from src.landmarker.heatmap.generator import GaussianHeatmapGenerator


def create_heatmap(
    subpixel=False, sigmas=1.0, rotations=0.0, return_covs=False, gamma=None, square=True
):
    """Create a heatmap and corresponding landmarks for testing."""
    # Create a generator with some example parameters
    if square:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 64),
            learnable=False,
            gamma=gamma,
        )

        # Create some example input data
        landmarks = torch.tensor([[[20, 20], [30, 40], [45, 25]]], dtype=torch.float)

        if subpixel:
            landmarks += torch.rand(landmarks.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks)

        if return_covs:
            return heatmaps, landmarks, generator.get_covariance_matrix()
        return heatmaps, landmarks
    else:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 32),
            learnable=False,
            gamma=gamma,
        )

        # Create some example input data
        landmarks = torch.tensor([[[20, 20], [30, 15], [45, 10]]], dtype=torch.float)

        if subpixel:
            landmarks += torch.rand(landmarks.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks)

        if return_covs:
            return heatmaps, landmarks, generator.get_covariance_matrix()
        return heatmaps, landmarks


def create_3d_heatmap(
    subpixel=False, sigmas=1.0, rotations=0.0, return_covs=False, gamma=None, square=True
):
    """Create a heatmap and corresponding landmarks for testing."""
    # Create a generator with some example parameters
    if square:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 64, 64),
            learnable=False,
            gamma=gamma,
        )

        # Create some example input data
        landmarks = torch.tensor([[[20, 20, 20], [30, 40, 30], [45, 25, 40]]], dtype=torch.float)

        if subpixel:
            landmarks += torch.rand(landmarks.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks)

        if return_covs:
            return heatmaps, landmarks, generator.get_covariance_matrix()
        return heatmaps, landmarks
    else:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 32, 40),
            learnable=False,
            gamma=gamma,
        )

        # Create some example input data
        landmarks = torch.tensor([[[20, 20, 20], [30, 15, 10], [45, 10, 15]]], dtype=torch.float)

        if subpixel:
            landmarks += torch.rand(landmarks.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks)

        if return_covs:
            return heatmaps, landmarks, generator.get_covariance_matrix()
        return heatmaps, landmarks


def create_batch_of_heatmaps(
    subpixel=False, sigmas=1.0, rotations=0.0, return_covs=False, gamma=None, square=True
):
    """Create a batch of heatmaps and corresponding landmarks for testing."""
    # Create a generator with some example parameters
    if square:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 64),
            learnable=False,
            gamma=gamma,
        )

        landmarks_batch = torch.tensor(
            [
                [[10, 10], [30, 40], [50, 60]],
                [[16, 17], [24, 35], [30, 58]],
                [[24, 24], [14, 14], [54, 30]],
                [[30, 30], [36, 46], [20, 10]],
            ],
            dtype=torch.float,
        )

        if subpixel:
            landmarks_batch += torch.rand(landmarks_batch.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks_batch)

        if return_covs:
            return heatmaps, landmarks_batch, generator.get_covariance_matrix()
        return heatmaps, landmarks_batch
    else:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 32),
            learnable=False,
            gamma=gamma,
        )

        landmarks_batch = torch.tensor(
            [
                [[10, 10], [30, 15], [50, 20]],
                [[16, 17], [24, 10], [30, 13]],
                [[24, 24], [14, 14], [54, 21]],
                [[30, 20], [36, 25], [20, 24]],
            ],
            dtype=torch.float,
        )

        if subpixel:
            landmarks_batch += torch.rand(landmarks_batch.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks_batch)

        if return_covs:
            return heatmaps, landmarks_batch, generator.get_covariance_matrix()
        return heatmaps, landmarks_batch


def create_batch_of_3d_heatmaps(
    subpixel=False, sigmas=1.0, rotations=0.0, return_covs=False, gamma=None, square=True
):
    """Create a batch of heatmaps and corresponding landmarks for testing."""
    # Create a generator with some example parameters
    if square:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 64, 64),
            learnable=False,
            gamma=gamma,
        )

        landmarks_batch = torch.tensor(
            [
                [[10, 10, 10], [30, 40, 30], [50, 60, 50]],
                [[16, 17, 16], [24, 35, 24], [30, 58, 30]],
                [[24, 24, 24], [14, 14, 14], [54, 30, 54]],
                [[30, 30, 30], [36, 46, 36], [20, 10, 20]],
            ],
            dtype=torch.float,
        )

        if subpixel:
            landmarks_batch += torch.rand(landmarks_batch.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks_batch)

        if return_covs:
            return heatmaps, landmarks_batch, generator.get_covariance_matrix()
        return heatmaps, landmarks_batch
    else:
        generator = GaussianHeatmapGenerator(
            nb_landmarks=3,
            sigmas=sigmas,
            rotation=rotations,
            heatmap_size=(64, 32, 40),
            learnable=False,
            gamma=gamma,
        )

        landmarks_batch = torch.tensor(
            [
                [[10, 10, 10], [30, 15, 20], [50, 20, 30]],
                [[16, 17, 16], [24, 10, 15], [30, 13, 10]],
                [[24, 24, 24], [14, 14, 14], [54, 21, 24]],
                [[30, 20, 30], [36, 25, 20], [20, 24, 10]],
            ],
            dtype=torch.float,
        )

        if subpixel:
            landmarks_batch += torch.rand(landmarks_batch.shape, dtype=torch.float32)

        # Call the method being tested
        heatmaps = generator(landmarks_batch)

        if return_covs:
            return heatmaps, landmarks_batch, generator.get_covariance_matrix()
        return heatmaps, landmarks_batch


def check_retriever_output(retriever_fun, atol=1e-2, rtol=0, spatial_dims=2, **kwargs):
    """Check the output of a retriever function."""
    # Create some example input data
    for square in [True, False]:
        if spatial_dims == 2:
            heatmaps, landmarks = create_heatmap(  # pylint: disable=unbalanced-tuple-unpacking
                subpixel=False, return_covs=False, square=square
            )
        else:
            heatmaps, landmarks = create_3d_heatmap(  # pylint: disable=unbalanced-tuple-unpacking
                subpixel=False, return_covs=False, square=square
            )

        # Call the method being tested
        coords = retriever_fun(heatmaps, spatial_dims=spatial_dims, **kwargs).float()

        # Check the output
        assert coords.shape == (1, 3, spatial_dims)
        assert torch.allclose(coords, landmarks, atol=atol, rtol=rtol)

        # Create some batched example input data
        if spatial_dims == 2:
            heatmaps, landmarks = (
                create_batch_of_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
                    subpixel=False, return_covs=False, square=square
                )
            )
        else:
            heatmaps, landmarks = (
                create_batch_of_3d_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
                    subpixel=False, return_covs=False, square=square
                )
            )
        # Call the method being tested
        coords = retriever_fun(heatmaps, spatial_dims=spatial_dims, **kwargs).float()

        # Check the output
        assert coords.shape == (4, 3, spatial_dims)
        assert torch.allclose(coords, landmarks, atol=atol, rtol=rtol)

        # Create some batched example input data with subpixel landmarks
        if spatial_dims == 2:
            heatmaps, landmarks = create_batch_of_heatmaps(
                subpixel=True, square=square
            )  # pylint: disable=unbalanced-tuple-unpacking
        else:
            heatmaps, landmarks = create_batch_of_3d_heatmaps(
                subpixel=True, square=square
            )  # pylint: disable=unbalanced-tuple-unpacking

        # Call the method being tested
        coords = retriever_fun(heatmaps, spatial_dims=spatial_dims, **kwargs).float()

        # Check the output
        assert coords.shape == (4, 3, spatial_dims)
        assert torch.allclose(coords.round(), landmarks.round(), atol=atol)


def test_coord_argmax():
    """Test the coord_argmax function."""
    check_retriever_output(coord_argmax, atol=1e-2, rtol=0)
    check_retriever_output(coord_argmax, atol=1e-2, rtol=0, spatial_dims=3)
    try:
        check_retriever_output(coord_argmax, atol=1e-2, rtol=0, spatial_dims=4)
        assert False
    except ValueError:
        assert True


def test_coord_local_soft_argmax():
    """Test the coord_local_soft_argmax function."""
    check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=3, t=10)

    check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=5, t=5)

    check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=7, t=3)
    check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=3, t=10, spatial_dims=3)

    check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=5, t=5, spatial_dims=3)

    check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=7, t=3, spatial_dims=3)
    try:
        check_retriever_output(coord_local_soft_argmax, atol=1e-2, window=7, t=3, spatial_dims=4)
        assert False
    except ValueError:
        assert True


def test_coord_weighted_spatial_mean():
    """Test the coord_weighted_spatial_mean function."""
    # TODO: the method seems to be kind of unstable hence the high tolerance
    check_retriever_output(coord_weighted_spatial_mean, atol=3)
    check_retriever_output(coord_weighted_spatial_mean, atol=3, spatial_dims=3)
    try:
        check_retriever_output(coord_weighted_spatial_mean, atol=3, spatial_dims=4)
        assert False
    except ValueError:
        assert True
    try:
        retriever_func = (
            lambda x, **kwargs: coord_weighted_spatial_mean(  # pylint: disable=unnecessary-lambda
                x, **kwargs, activation="wrong"
            )
        )
        check_retriever_output(retriever_func, atol=3)
        assert False
    except ValueError:
        assert True


def test_coord_cov_from_guassian_ls_scipy():
    """Test the coord_cov_from_guassian_ls_scipy function."""
    # Create some example input data
    rotations = np.array([0, 0, 0])
    sigmas = np.array([[5, 5], [1, 1], [2, 1]])
    heatmaps, landmarks, covs_true = create_heatmap(
        subpixel=False, sigmas=sigmas, rotations=rotations, return_covs=True
    )
    covs_true = covs_true.unsqueeze(0)
    # Call the method being tested
    coords, covs = coord_cov_from_gaussian_ls(heatmaps, gamma=None, ls_library="scipy")
    coords = coords.float()

    # Check the output
    assert coords.shape == (1, 3, 2)
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5)

    # Create some batched example input data
    heatmaps, landmarks = create_batch_of_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
        subpixel=False
    )

    # Call the method being tested
    coords, covs = coord_cov_from_gaussian_ls(heatmaps, gamma=None, ls_library="scipy")
    coords = coords.float()

    # Check the output
    assert coords.shape == (4, 3, 2)
    assert covs.shape == (4, 3, 2, 2)
    assert torch.allclose(coords, landmarks, atol=0.5)

    # Create some batched example input data with subpixel landmarks
    heatmaps, landmarks, covs_true = create_batch_of_heatmaps(
        subpixel=True, sigmas=sigmas, rotations=rotations, return_covs=True
    )
    coords = coords.float()

    # Call the method being tested
    coords, covs = coord_cov_from_gaussian_ls(heatmaps, gamma=None, ls_library="scipy")
    coords = coords.float()

    # Check the output
    assert coords.shape == (4, 3, 2)
    assert covs.shape == (4, 3, 2, 2)
    assert torch.allclose(coords.round(), landmarks.round(), atol=0.5)
    assert torch.allclose(covs, covs_true, atol=0.5)

    try:
        _ = coord_cov_from_gaussian_ls(heatmaps, gamma=None, ls_library="scipy", spatial_dims=3)
        assert False
    except ValueError:
        assert True

    try:
        _ = coord_cov_from_gaussian_ls(
            heatmaps, gamma=None, ls_library="Not implemented", spatial_dims=2
        )
        assert False
    except ValueError:
        assert True


def test_coord_cov_from_weighted_spatial_mean():
    """Test the coord_cov_from_weighted_spatial_mean function."""
    # Create some example input data
    rotations = np.array([0, 0, 0])
    sigmas = np.array([[1, 1], [1, 1], [1, 1]])

    h, w = 128, 128

    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3, sigmas=sigmas, rotation=rotations, heatmap_size=(h, w), learnable=False
    )
    landmarks = torch.tensor(
        [[[h // 2, w // 2], [h // 2, w // 2], [h // 2, w // 2]]], dtype=torch.float
    )
    heatmaps = heatmap_generator(landmarks)

    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean")
    coords = coords.float()

    # Check the output
    assert coords.shape == (1, 3, 2)
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(coords, landmarks, atol=0.5)
    assert torch.allclose(covs, covs_true, atol=0.5)

    # Create some example input data
    rotations = np.array([0, 0, 0])
    sigmas = np.array([[5, 5], [1, 1], [2, 1]])
    heatmaps, landmarks, covs_true = create_heatmap(  # pylint: disable=unbalanced-tuple-unpacking
        subpixel=False, sigmas=sigmas, rotations=rotations, return_covs=True, gamma=None
    )
    covs_true = covs_true.unsqueeze(0)
    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean")
    coords = coords.float()

    # Check the output
    assert coords.shape == (1, 3, 2)
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=2)

    # Create some batched example input data
    heatmaps, landmarks = create_batch_of_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
        subpixel=False
    )

    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean")
    coords = coords.float()

    # Check the output
    assert coords.shape == (4, 3, 2)
    assert covs.shape == (4, 3, 2, 2)
    assert torch.allclose(coords, landmarks, atol=0.5)

    # Create some batched example input data with subpixel landmarks
    heatmaps, landmarks, covs_true = create_batch_of_heatmaps(
        subpixel=True, sigmas=sigmas, rotations=rotations, return_covs=True
    )
    coords = coords.float()

    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean")
    coords = coords.float()

    # Check the output
    assert coords.shape == (4, 3, 2)
    assert covs.shape == (4, 3, 2, 2)
    assert torch.allclose(coords, landmarks, atol=0.5)
    assert torch.allclose(covs, covs_true, atol=0.5, rtol=0.1)


def test_coord_cov_from_weighted_spatial_mean_3d():
    """Test the coord_cov_from_weighted_spatial_mean function."""
    # Create some example input data
    rotations = np.array([0, 0, 0])
    sigmas = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])

    h, w, d = 128, 128, 128

    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3, sigmas=sigmas, rotation=rotations, heatmap_size=(h, w, d), learnable=False
    )
    landmarks = torch.tensor(
        [[[h // 2, w // 2, d // 2], [h // 2, w // 2, d // 2], [h // 2, w // 2, d // 2]]],
        dtype=torch.float,
    )
    heatmaps = heatmap_generator(landmarks)

    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean", spatial_dims=3)
    coords = coords.float()

    # Check the output
    assert coords.shape == (1, 3, 3)
    assert covs.shape == (1, 3, 3, 3)
    assert torch.allclose(coords, landmarks, atol=0.5)
    assert torch.allclose(covs, covs_true, atol=0.5)

    # Create some example input data
    rotations = np.array([0, 0, 0])
    sigmas = np.array([[5, 5, 3], [1, 1, 2], [2, 1, 3]])
    heatmaps, landmarks, covs_true = (
        create_3d_heatmap(  # pylint: disable=unbalanced-tuple-unpacking
            subpixel=False, sigmas=sigmas, rotations=rotations, return_covs=True, gamma=None
        )
    )
    covs_true = covs_true.unsqueeze(0)
    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean", spatial_dims=3)
    coords = coords.float()

    # Check the output
    assert coords.shape == (1, 3, 3)
    assert covs.shape == (1, 3, 3, 3)
    assert torch.allclose(covs, covs_true, atol=2)

    # Create some batched example input data
    heatmaps, landmarks = create_batch_of_3d_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
        subpixel=False
    )

    # Call the method being tested
    coords, covs = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean", spatial_dims=3)
    coords = coords.float()

    # Check the output
    assert coords.shape == (4, 3, 3)
    assert covs.shape == (4, 3, 3, 3)
    assert torch.allclose(coords, landmarks, atol=0.5)

    try:
        _ = heatmap_to_coord_cov(heatmaps, method="weighted_spatial_mean", spatial_dims=4)
        assert False
    except ValueError:
        assert True


def test_windowed_weigthed_sample_cov():
    """Test the windowed_weigthed_sample_cov function."""
    # Create some example input data
    rotations = np.array([0, 0, 0])
    sigmas = np.array([[3, 3], [1, 1], [2, 1]])

    h, w = 512, 512
    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3, sigmas=sigmas, rotation=rotations, heatmap_size=(h, w), learnable=False
    )
    landmarks = torch.tensor(
        [[[h // 2, w // 2], [h // 2, w // 2], [h // 2, w // 2]]], dtype=torch.float
    )
    heatmaps = heatmap_generator(landmarks)

    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    covs = windowed_weigthed_sample_cov(heatmaps, landmarks)

    # Check the output
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5, rtol=0.1)

    heatmaps, landmarks, covs_true = create_heatmap(
        subpixel=False, sigmas=sigmas, rotations=rotations, return_covs=True, gamma=None
    )
    covs_true = covs_true.unsqueeze(0)
    # Call the method being tested
    covs = windowed_weigthed_sample_cov(heatmaps, landmarks)

    # Check the output
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5, rtol=0.1)


def test_cov_ls_scipy():
    """Test cov_ls_scipy."""
    # Create some example input data
    rotations = np.array([0.0, 0, 1])
    sigmas = np.array([[5, 5], [1, 1], [2, 1]])

    gamma = None
    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3,
        sigmas=sigmas,
        rotation=rotations,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=gamma,
    )
    landmarks = torch.tensor(
        [[[64 // 2 - 5, 64 // 2], [64 // 2, 64 // 2 + 5], [64 // 2, 64 // 2]]], dtype=torch.float
    )
    heatmaps = heatmap_generator(landmarks)
    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    covs = cov_from_gaussian_ls(heatmaps, landmarks, gamma=gamma, ls_library="scipy")

    # Check the output
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5)

    gamma = 1
    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3,
        sigmas=sigmas,
        rotation=rotations,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=gamma,
    )
    landmarks = torch.tensor(
        [[[64 // 2 - 5, 64 // 2], [64 // 2, 64 // 2 + 5], [64 // 2, 64 // 2]]], dtype=torch.float
    )
    heatmaps = heatmap_generator(landmarks)
    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    covs = cov_from_gaussian_ls(heatmaps, landmarks, gamma=gamma)

    # Check the output
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5)


def test_weighted_sample_cov():
    """Test weighted_sample_cov."""
    # Create some example input data
    rotations = np.array([0.0, 0, 1])
    sigmas = np.array([[5, 5], [1, 1], [2, 1]])

    gamma = 1
    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3,
        sigmas=sigmas,
        rotation=rotations,
        heatmap_size=(64, 64),
        learnable=False,
        gamma=gamma,
    )
    landmarks = torch.tensor(
        [[[64 // 2 - 5, 64 // 2], [64 // 2, 64 // 2 + 5], [64 // 2, 64 // 2]]], dtype=torch.float
    )
    heatmaps = heatmap_generator(landmarks)
    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    covs = weighted_sample_cov(heatmaps, landmarks, spatial_dims=2)

    # Check the output
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5)

    # Call the method being tested
    covs = weighted_sample_cov(torch.logit(heatmaps, eps=1e-38), landmarks, activation="sigmoid")
    # Check the output
    assert covs.shape == (1, 3, 2, 2)
    assert torch.allclose(covs, covs_true, atol=0.5)


def test_weighted_sample_cov_3d():
    """Test weighted_sample_cov for 3D."""
    # Create some example input data
    rotations = np.array([0.0, 0, 1])
    sigmas = np.array([[4, 4, 5], [1, 1, 1], [2, 1, 1]])

    gamma = 1
    heatmap_generator = GaussianHeatmapGenerator(
        nb_landmarks=3,
        sigmas=sigmas,
        rotation=rotations,
        heatmap_size=(64, 64, 64),
        learnable=False,
        gamma=gamma,
    )
    landmarks = torch.tensor(
        [
            [
                [64 // 2 - 5, 64 // 2, 64 // 2],
                [64 // 2, 64 // 2 + 5, 64 // 2],
                [64 // 2, 64 // 2, 64 // 2 - 3],
            ]
        ],
        dtype=torch.float,
    )
    heatmaps = heatmap_generator(landmarks)
    covs_true = heatmap_generator.get_covariance_matrix().unsqueeze(0)

    # Call the method being tested
    covs = weighted_sample_cov(heatmaps, landmarks, spatial_dims=3)

    # Check the output
    assert covs.shape == (1, 3, 3, 3)
    assert torch.allclose(covs, covs_true, atol=0.5)

    # Call the method being tested
    covs = weighted_sample_cov(
        torch.logit(heatmaps, eps=1e-38), landmarks, activation="sigmoid", spatial_dims=3
    )
    # Check the output
    assert covs.shape == (1, 3, 3, 3)
    assert torch.allclose(covs, covs_true, atol=0.5)


def test_heatmap_to_coord():
    """Test the heatmap to coord function."""
    heatmap, _ = create_batch_of_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
        subpixel=False
    )

    coord_1 = coord_argmax(heatmap)
    coord_2 = heatmap_to_coord(heatmap, method="argmax")
    assert torch.allclose(coord_1, coord_2)

    coord_1 = coord_local_soft_argmax(heatmap)
    coord_2 = heatmap_to_coord(heatmap, method="local_soft_argmax")
    assert torch.allclose(coord_1, coord_2)

    coord_1 = coord_weighted_spatial_mean(heatmap)
    coord_2 = heatmap_to_coord(heatmap, method="weighted_spatial_mean")
    assert torch.allclose(coord_1, coord_2)

    coord_1 = coord_weighted_spatial_mean(heatmap, activation="softmax")
    coord_2 = heatmap_to_coord(heatmap, method="soft_argmax")
    assert torch.allclose(coord_1, coord_2)

    coord_1 = coord_weighted_spatial_mean(heatmap, activation="ReLU")
    coord_2 = heatmap_to_coord(heatmap, method="weighted_spatial_mean_relu")

    coord_1 = coord_weighted_spatial_mean(heatmap, activation="sigmoid")
    coord_2 = heatmap_to_coord(heatmap, method="weighted_spatial_mean_sigmoid")

    try:
        _ = heatmap_to_coord(heatmap, method="wrong_method")
        assert False
    except ValueError:
        assert True


def test_heatmap_to_multiple_coord():
    """Test the heatmap to coord function."""
    heatmap, landmark = create_batch_of_heatmaps(  # pylint: disable=unbalanced-tuple-unpacking
        subpixel=False
    )
    heatmap = heatmap.sum(dim=0).unsqueeze(0)
    assert heatmap.shape == (1, 3, 64, 64)

    pred_coords, _ = heatmap_to_multiple_coord(heatmap, window=5, threshold=0.5, method="argmax")
    assert len(pred_coords) == 1
    for i in range(1):
        assert len(pred_coords[i]) == 3
        for j in range(3):
            assert len(pred_coords[i][j]) == 4
            for k in range(4):
                assert any([torch.allclose(pred_coords[i][j][k], landmark[b, j]) for b in range(4)])


def test_heatmap_to_coord_enlarge():
    """Test the heatmap to coord function."""
    check_retriever_output(
        heatmap_to_coord_enlarge, atol=1, rtol=0.1, method="argmax", enlarge_factor=2
    )

    check_retriever_output(
        heatmap_to_coord_enlarge, atol=1, rtol=0.1, method="argmax", enlarge_dim=(128, 128)
    )
    check_retriever_output(
        heatmap_to_coord_enlarge,
        atol=1,
        rtol=0.1,
        method="argmax",
        enlarge_factor=2,
        spatial_dims=3,
        enlarge_mode="trilinear",
    )
    check_retriever_output(
        heatmap_to_coord_enlarge,
        atol=1,
        rtol=0.1,
        method="argmax",
        enlarge_dim=(128, 128, 128),
        spatial_dims=3,
        enlarge_mode="trilinear",
    )
    try:
        check_retriever_output(
            heatmap_to_coord_enlarge,
            atol=1,
            rtol=0.1,
            method="argmax",
            enlarge_dim=(128, 128),
            spatial_dims=5,
        )
        assert False
    except ValueError:
        assert True



"""Module containing the ``LandmarkDataset`` class and its subclasses ``MaskDataset``
and ``HeatmapDataset``. All three classe allow to create a dataset of images and landmarks.
The three classes are subclasses of ``torch.utils.data.Dataset``.
"""

from typing import Callable, Optional

import numpy as np
import torch
import torch.nn.functional as F
from monai.transforms import Compose, LoadImage, Resize, Transpose  # type: ignore
from torch.utils.data import Dataset
from tqdm import tqdm  # type: ignore

from landmarker.heatmap.generator import (
    GaussianHeatmapGenerator,
    HeatmapGenerator,
    LaplacianHeatmapGenerator,
)
from landmarker.transforms.images import resize_with_pad
from landmarker.transforms.landmarks import affine_landmarks, resize_landmarks
from landmarker.utils import extract_roi


class LandmarkDataset(Dataset):
    """
    ``LandmarkDataset`` is a subclass of ``torch.utils.data.Dataset``. It represents a dataset of
    images and landmarks. The images can be provided as a list of paths to the images or as a list
    of numpy arrays or as a numpy array/torch.Tensor.

    Args:
        imgs (list[str] | list[np.array] | np.ndarray | torch.Tensor): list of paths to the images
            or list of numpy arrays or numpy array/torch.Tensor.
        landmarks (np.ndarray | torch.Tensor): landmarks of the images. If the landmarks are 2D: y, x
            or 3D: z, y, x.
        spatial_dims (int): number of spatial dimensions of the images. (defaults: 2)
        pixel_spacing (Optional[np.ndarray | torch.Tensor]): pixel spacing of the images.
            (defaults: None)
        class_names (Optional[list]): names of the landmarks. (defaults: None)
        transform (Optional[Callable]): transformation to apply to the images and landmarks.
            (defaults: None)
        store_imgs (bool): whether to store the images in memory or not. (defaults: True)
        dim_img (Optional[tuple[int, int]]): dimension of the images. (defaults: None)
        img_paths (Optional[list[str]]): list of paths to the images. (defaults: None)
        resize_pad (bool): whether to resize the images and landmarks or not. (defaults: True)
    """

    def __init__(
        self,
        imgs: list[str] | np.ndarray | torch.Tensor | list[np.ndarray] | list[torch.Tensor],
        landmarks: np.ndarray | torch.Tensor,
        spatial_dims: int = 2,
        pixel_spacing: Optional[list[int] | tuple[int, ...] | np.ndarray | torch.Tensor] = None,
        class_names: Optional[list] = None,
        transform: Optional[Callable] = None,
        store_imgs: bool = True,
        dim_img: Optional[tuple[int, ...]] = None,
        img_paths: Optional[list[str]] = None,
        resize_pad: bool = True,
    ) -> None:
        if spatial_dims not in [2, 3]:
            raise ValueError(f"spatial_dims must be 2 or 3, got {spatial_dims}")
        self.spatial_dims = spatial_dims
        self.resize_pad = resize_pad
        self.transform = transform
        self.store_imgs = store_imgs
        self.dim_img = dim_img
        if isinstance(imgs, list) and isinstance(imgs[0], str):
            if imgs[0].endswith(".npy"):
                self.image_loader = Compose(
                    [
                        LoadImage(image_only=True, ensure_channel_first=True),
                    ]
                )
            elif self.spatial_dims == 2:
                self.image_loader = Compose(
                    [
                        LoadImage(image_only=True, ensure_channel_first=True),
                        Transpose(indices=[0, 2, 1]),
                    ]
                )
            else:
                self.image_loader = Compose(
                    [
                        LoadImage(image_only=True, ensure_channel_first=True),
                        Transpose(indices=[0, 3, 2, 1]),
                    ]
                )
        self._init_landmarks(landmarks, class_names)
        self._init_imgs(imgs, img_paths)
        self._init_pixel_spacing(pixel_spacing)

    def _init_landmarks(
        self,
        landmarks: np.ndarray | torch.Tensor,
        class_names: Optional[list] = None,
    ) -> None:
        if len(landmarks.shape) == 2:
            landmarks = landmarks.reshape(
                (landmarks.shape[0], 1, landmarks.shape[1])
            )  # (N, D) => (N, 1, D) only one class
        self.nb_landmarks = landmarks.shape[1]
        assert self.spatial_dims == landmarks.shape[-1]
        if isinstance(landmarks, np.ndarray):
            self.landmarks_original = torch.Tensor(landmarks.copy())
        elif isinstance(landmarks, torch.Tensor):
            self.landmarks_original = landmarks.clone().float()
        else:
            if landmarks is None:
                raise ValueError("landmarks must be provided")
            raise TypeError("landmarks type not supported")

        if class_names is None:
            self.class_names = [f"landmark_{i}" for i in range(self.landmarks_original.shape[1])]
        else:
            if len(class_names) != self.landmarks_original.shape[1]:
                raise ValueError("class_names must be of length landmarks.shape[1] or None")
            self.class_names = class_names

    def _init_pixel_spacing(
        self, pixel_spacing: Optional[list[int] | tuple[int, ...] | np.ndarray | torch.Tensor]
    ) -> None:
        if pixel_spacing is not None:
            if isinstance(pixel_spacing, (tuple, list)):
                assert (
                    len(pixel_spacing) == self.spatial_dims
                ), f"pixel_spacing must be of shape {self.spatial_dims}"
                self.pixel_spacings = (
                    torch.Tensor(pixel_spacing).unsqueeze(0).repeat(len(self.landmarks_original), 1)
                )
            elif pixel_spacing.shape == (len(self.landmarks_original), self.spatial_dims):
                self.pixel_spacings = torch.Tensor(pixel_spacing)
            elif pixel_spacing.shape == (self.spatial_dims,):
                self.pixel_spacings = (
                    torch.Tensor(pixel_spacing).unsqueeze(0).repeat(len(self.landmarks_original), 1)
                )
            elif pixel_spacing.shape == (len(self.landmarks_original),):
                self.pixel_spacings = (
                    torch.Tensor(pixel_spacing).unsqueeze(1).repeat(1, self.spatial_dims)
                )
            else:
                raise ValueError(
                    f"Pixel_spacing must be of shape (N, {self.spatial_dims}) or "
                    "({self.spatial_dims},) since spatial_dims is {self.spatial_dims}."
                )
        else:
            self.pixel_spacings = torch.ones(len(self.landmarks_original), self.spatial_dims)

    def _init_imgs(
        self,
        imgs: list[str] | np.ndarray | torch.Tensor | list[np.ndarray] | list[torch.Tensor],
        img_paths: Optional[list[str]],
    ) -> None:
        if isinstance(imgs, list):
            if isinstance(imgs[0], str):
                self.img_paths = imgs
            elif isinstance(imgs[0], np.ndarray):
                self.imgs = [torch.Tensor(imgs[i]) for i in range(len(imgs))]
            elif isinstance(imgs[0], torch.Tensor):
                self.imgs = imgs  # type: ignore
            else:
                raise TypeError(f"imgs type not supported, got {type(imgs[0])}")
        else:
            self.imgs = [torch.Tensor(imgs[i]) for i in range(len(imgs))]
            self.img_paths = img_paths if img_paths is not None else []
        if not isinstance(imgs[0], str):
            if self.spatial_dims == 2:
                assert len(self.imgs[0].shape) == 3
            else:
                assert len(self.imgs[0].shape) == 4

        # images need to be stored in memory and images are provided as paths
        if self.store_imgs and not hasattr(self, "imgs"):
            assert self.img_paths is not None
            self.imgs, self.dim_original = self._read_all_imgs(self.img_paths)  # type: ignore
        if self.dim_img is not None and self.store_imgs:
            (
                self.imgs,
                self.landmarks,
                self.dim_original,
                self.paddings,
            ) = self._resize_all_imgs_landmarks(self.imgs, self.landmarks_original, self.dim_img)
        else:
            self.landmarks = self.landmarks_original.clone()
            self.paddings = torch.zeros((self.landmarks.shape[0], self.spatial_dims))
            if not hasattr(self, "dim_original") and self.store_imgs:
                self.dim_original = torch.zeros((self.landmarks.shape[0], self.spatial_dims))
                for i in range(self.landmarks.shape[0]):
                    self.dim_original[i] = torch.tensor(
                        self.imgs[i].shape[-self.spatial_dims :]
                    ).float()

    def __len__(self) -> int:
        return len(self.landmarks_original)

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        if not self.store_imgs:
            if not hasattr(self, "image_loader"):
                img = self.imgs[idx]  # type: ignore
            else:
                img = self.image_loader(self.img_paths[idx])  # type: ignore
            dim_original = torch.tensor(img.shape[-self.spatial_dims :]).float()
            if self.dim_img is not None:
                if self.resize_pad:
                    dim_original = torch.tensor(img.shape[-self.spatial_dims :]).float()
                    img, pad = resize_with_pad(img, self.dim_img)
                    padding = torch.Tensor(pad)
                    landmark = resize_landmarks(
                        self.landmarks_original[idx], dim_original, self.dim_img, padding
                    )
                else:
                    img = Resize(spatial_size=self.dim_img)(img)
                    landmark = resize_landmarks(
                        self.landmarks_original[idx], dim_original, self.dim_img
                    )
                    padding = torch.zeros(self.spatial_dims)
            else:
                landmark = self.landmarks_original[idx]
                padding = torch.zeros(self.spatial_dims)
        else:
            img = self.imgs[idx]
            landmark = self.landmarks[idx]
            padding = self.paddings[idx]
            dim_original = self.dim_original[idx]
        if self.transform is not None:
            img, landmark_t, affine_matrix = self._transform_img_landmark(
                img, landmark, self.transform  # type: ignore
            )
        else:
            landmark_t = landmark
            affine_matrix = torch.eye(4)

        return {
            "image": img,
            "landmark": landmark_t,
            "affine": affine_matrix,
            "dim_original": dim_original,
            "spacing": self.pixel_spacings[idx],
            "padding": padding,
        }

    def _transform_img_landmark(
        self, img: torch.Tensor, landmark: torch.Tensor, transform: Compose
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Transform the image and the landmark using the transformation arg transform.

        Args:
            img (torch.Tensor): image to transform.
            landmark (torch.Tensor): landmark to transform.
            transform (Callable): transformation to apply to the image and the landmark.

        Returns:
            img_t (torch.Tensor): transformed image.
            landmark_t (torch.Tensor): transformed landmark.
            affine_matrix_push (torch.Tensor): push affine matrix of the transformation.
        """
        img_t = transform({"image": img})["image"]  # type: ignore
        affine_matrix_pull = img_t.meta.get("affine", None).float()  # type: ignore
        affine_matrix_push = torch.linalg.inv(affine_matrix_pull)
        landmark_t = affine_landmarks(landmark, affine_matrix_push)
        return img_t, landmark_t, affine_matrix_push

    def _read_all_imgs(self, paths: list[str]) -> tuple[list[torch.Tensor], torch.Tensor]:
        """
        Read all images in ``paths``.

        Args:
            paths (list[str]): list of paths to the images.

        Returns:
            imgs (list[torch.Tensor]): list of images.
            dim_original (torch.Tensor): original dimensions of the images.
        """
        print(f"Reading {len(paths)} images...")
        imgs = []
        dim_original = torch.zeros((len(paths), self.spatial_dims))
        for i, path in enumerate(tqdm(paths)):
            img = self.image_loader(path)
            dim_original[i] = torch.tensor(img.shape[-self.spatial_dims :]).float()
            imgs.append(img)
        return imgs, dim_original

    def _resize_all_imgs_landmarks(
        self, imgs: list[torch.Tensor], landmarks: torch.Tensor, dim_img: tuple[int, ...]
    ) -> tuple[list[torch.Tensor], torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Resize all images in ``imgs`` and resize all landmarks in ``landmarks``.

        Args:
            imgs (torch.Tensor): images.
            landmarks (torch.Tensor): landmarks of the images.
            dim_img (tuple[int, int]): dimension of the images.

        Returns:
            imgs (list[torch.Tensor]): list of resized images.
            landmarks (torch.Tensor): resized landmarks.
            dim_original (torch.Tensor): original dimensions of the images.
            paddings (torch.Tensor): paddings applied to the images.
        """
        print(f"Resizing {len(imgs)} images and landmarks...")
        landmarks = landmarks.clone()
        dim_original = torch.zeros((len(imgs), self.spatial_dims))
        paddings = torch.zeros((len(imgs), self.spatial_dims))
        for i in tqdm(range(len(imgs))):
            dim_original[i] = torch.tensor(imgs[i].shape[-self.spatial_dims :]).float()
            if self.resize_pad:
                imgs[i], padding = resize_with_pad(imgs[i], dim_img)
                paddings[i] = torch.Tensor(padding)
            else:
                imgs[i] = Resize(dim_img)(imgs[i])
                paddings[i] = torch.zeros(self.spatial_dims)
            landmarks[i] = resize_landmarks(landmarks[i], dim_original[i], dim_img, paddings[i])
        return imgs, landmarks, dim_original, paddings


class PatchDataset(LandmarkDataset):
    """
    ``PatchDataset`` is a subclass of ``LandmarkDataset``. It represents a dataset of images and
    landmarks. The images can be provided as a list of paths to the images or as a list of numpy
    arrays or as a numpy array/torch.Tensor. The landmarks can be provided as a numpy array or
    torch.Tensor. The landmarks can be extracted from the images and stored as a torch.Tensor or
    they can be genererated from provided landmarks. The images are cropped around the landmarks
    and stored as torch.Tensor. The landmarks are the landmarks of the cropped images. The offset
    is randomly generated in a range of [-range_aug_patch, range_aug_patch] and added to the
    landmarks to create the cropped images. If the cropped images are not of the same size, they
    are resized to the same size.

    Args:
        imgs (list[str] | list[np.array] | np.ndarray | torch.Tensor): list of paths to the images
            or list of numpy arrays or numpy array/torch.Tensor.
        landmarks (np.ndarray | torch.Tensor): landmarks of the images. If the landmarks are 2D: y, x
            or 3D: z, y, x.
        index_landmark (int): index of the landmark to use to extract the patch. (defaults: 0)
        spatial_dims (int): number of spatial dimensions of the images. (defaults: 2)
        pixel_spacing (Optional[np.ndarray | torch.Tensor]): pixel spacing of the images.
            (defaults: None)
        class_names (Optional[list]): names of the landmarks. (defaults: None)
        transform (Optional[Callable]): transformation to apply to the images and landmarks.
            (defaults: None)
        store_imgs (bool): whether to store the images in memory or not. (defaults: True)
        dim_patch (tuple[int, ...]): dimension of the patches. (defaults: (256, 256))
        range_aug_patch (int): range of the random offset to apply to the landmarks to extract the
            patches. (defaults: 64)

    """

    def __init__(
        self,
        imgs: list[str] | list[np.ndarray] | np.ndarray | torch.Tensor,
        landmarks: np.ndarray | torch.Tensor,
        index_landmark=0,
        spatial_dims: int = 2,
        pixel_spacing: Optional[torch.Tensor] = None,
        class_names: Optional[list] = None,
        transform: Optional[Callable] = None,
        store_imgs: bool = False,
        dim_patch: tuple[int, ...] = (256, 256),
        range_aug_patch: int = 64,
    ) -> None:
        self.range_aug_patch = range_aug_patch
        self.dim_patch = dim_patch
        self.index_landmark = index_landmark
        super().__init__(
            imgs=imgs,
            landmarks=landmarks,
            spatial_dims=spatial_dims,
            pixel_spacing=pixel_spacing,
            class_names=class_names,
            transform=transform,
            store_imgs=store_imgs,
            dim_img=None,
            resize_pad=False,
        )

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        results_landmark_ds = super().__getitem__(idx)
        img = results_landmark_ds["image"]
        landmark = results_landmark_ds["landmark"][self.index_landmark]
        if self.range_aug_patch is not None:
            offset = torch.randint(-self.range_aug_patch, self.range_aug_patch, landmark.shape)
            roi_middle = landmark + offset
        else:
            offset = torch.zeros(landmark.shape)
            roi_middle = landmark
        patch, landmark_patch, roi_corner_point = extract_roi(
            img, roi_middle, landmark, self.dim_patch, self.spatial_dims, ensure_dim=True
        )
        return {
            "image": patch,
            "landmark": landmark_patch,
            "offset": offset,
            "landmark_original": landmark,
            "roi_corner_point": roi_corner_point,
            "affine": results_landmark_ds["affine"],
            "dim_original": torch.tensor(self.dim_patch),
            "dim_original_full_image": results_landmark_ds["dim_original"],
            "spacing": results_landmark_ds["spacing"],
            "padding": results_landmark_ds["padding"],
        }


class MaskDataset(LandmarkDataset):
    """
    ``MaskDataset`` is a subclass of ``LandmarkDataset``. It represents a dataset of images
    and landmarks. The images can be provided as a list of paths to the images or as a list of
    numpy arrays or as a numpy array/torch.Tensor. The landmarks can be provided as a list of paths
    to the masks or as a list of numpy arrays or as a numpy array/torch.Tensor. The masks must be
    grayscale images with the landmarks represented by different values. The landmarks can be
    extracted from the masks and stored as a torch.Tensor or they can be genererated from provided
    landmarks. Note that if mask paths are provide this can be only be grayscale images and thus
    the dataset can only contain one class of landmark per image.

    Args:
        imgs (list[str] | list[np.array] | np.ndarray | torch.Tensor): list of paths to the images
            or list of numpy arrays or numpy array/torch.Tensor.
        landmarks (np.ndarray | torch.Tensor): landmarks of the images. If the landmarks are 2D: y, x
            or 3D: z, y, x.
        spatial_dims (int): number of spatial dimensions of the images. (defaults: 2)
        pixel_spacing (Optional[np.ndarray | torch.Tensor]): pixel spacing of the images.
            (defaults: None)
        class_names (Optional[list]): names of the landmarks. (defaults: None)
        transform (Optional[Callable]): transformation to apply to the images and landmarks.
            (defaults: None)
        store_imgs (bool): whether to store the images in memory or not. (defaults: True)
        store_masks (bool): whether to store the masks in memory or not.
            (defaults: False)
        dim_img (Optional[tuple[int, int]]): dimension of the images. (defaults: None)
        mask_paths (Optional[list[str]]): list of paths to the masks. (defaults: None)
        nb_landmarks (int): number of landmarks in the masks. (defaults: 1)
        img_paths (Optional[list[str]]): list of paths to the images. (defaults: None)
        resize_pad (bool): whether to resize the images and landmarks or not. (defaults: True)
    """

    def __init__(
        self,
        imgs: list[str] | list[np.ndarray] | np.ndarray | torch.Tensor,
        landmarks: Optional[np.ndarray | torch.Tensor] = None,
        spatial_dims: int = 2,
        pixel_spacing: Optional[torch.Tensor] = None,
        class_names: Optional[list] = None,
        transform: Optional[Callable] = None,
        store_imgs: bool = True,
        dim_img: Optional[tuple[int, ...]] = None,
        mask_paths: Optional[list[str]] = None,
        nb_landmarks: int = 1,
        img_paths: Optional[list[str]] = None,
        resize_pad: bool = True,
    ) -> None:
        self.nb_landmarks = nb_landmarks
        self.mask_paths = mask_paths
        self.dim_img = dim_img
        self.resize_pad = resize_pad
        self.spatial_dims = spatial_dims
        if mask_paths is not None:
            if mask_paths[0].endswith(".npy"):
                self.mask_loader = Compose(
                    [
                        LoadImage(image_only=True, ensure_channel_first=True),
                    ]
                )
            elif self.spatial_dims == 2:
                self.mask_loader = Compose(
                    [
                        LoadImage(image_only=True, ensure_channel_first=True),
                        Transpose(indices=[0, 2, 1]),
                    ]
                )
            else:
                self.mask_loader = Compose(
                    [
                        LoadImage(image_only=True, ensure_channel_first=True),
                        Transpose(indices=[0, 3, 2, 1]),
                    ]
                )
        if landmarks is not None:
            super().__init__(
                imgs=imgs,
                landmarks=landmarks,
                spatial_dims=spatial_dims,
                pixel_spacing=pixel_spacing,
                class_names=class_names,
                transform=transform,
                store_imgs=store_imgs,
                dim_img=dim_img,
                img_paths=img_paths,
                resize_pad=resize_pad,
            )
        elif self.mask_paths is not None:
            landmarks = self._read_extract_landmarks(self.mask_paths)
            super().__init__(
                imgs=imgs,
                landmarks=landmarks,
                spatial_dims=spatial_dims,
                pixel_spacing=pixel_spacing,
                class_names=class_names,
                transform=transform,
                store_imgs=store_imgs,
                dim_img=dim_img,
                img_paths=img_paths,
                resize_pad=resize_pad,
            )
        else:
            raise ValueError("Either mask_paths or landmarks must be provided")

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        if not self.store_imgs:
            if not hasattr(self, "image_loader"):
                img = self.imgs[idx]  # type: ignore
            else:
                img = self.image_loader(self.img_paths[idx])  # type: ignore
            dim_original = torch.tensor(img.shape[-self.spatial_dims :]).float()
            if self.dim_img is not None:
                if self.resize_pad:
                    dim_original = torch.tensor(img.shape[-self.spatial_dims :]).float()
                    img, pad = resize_with_pad(img, self.dim_img)
                    padding = torch.Tensor(pad)
                    landmark = resize_landmarks(
                        self.landmarks_original[idx], dim_original, self.dim_img, padding
                    )
                else:
                    img = Resize(self.dim_img)(img)
                    landmark = resize_landmarks(
                        self.landmarks_original[idx], dim_original, self.dim_img
                    )
                    padding = torch.zeros(self.spatial_dims)
                dim_img = self.dim_img
            else:
                landmark = self.landmarks_original[idx]
                padding = torch.zeros(self.spatial_dims)
                if self.spatial_dims == 2:
                    dim_img = (int(dim_original[0].item()), int(dim_original[1].item()))
                else:
                    dim_img = (
                        int(dim_original[0].item()),
                        int(dim_original[1].item()),
                        int(dim_original[2].item()),
                    )
        else:
            img = self.imgs[idx]
            landmark = self.landmarks[idx]
            padding = self.paddings[idx]
            dim_original = self.dim_original[idx]
            if self.dim_img is not None:
                dim_img = self.dim_img
            else:
                if self.spatial_dims == 2:
                    dim_img = (int(dim_original[0].item()), int(dim_original[1].item()))
                else:
                    dim_img = (
                        int(dim_original[0].item()),
                        int(dim_original[1].item()),
                        int(dim_original[2].item()),
                    )
        if self.transform is not None:
            img, landmark_t, affine_matrix = self._transform_img_landmark(
                img, landmark, self.transform  # type: ignore
            )
        else:
            landmark_t = landmark
            affine_matrix = torch.eye(4)
        mask = self._create_mask(landmark_t, dim_img)
        return {
            "image": img,
            "mask": mask,
            "landmark": landmark_t,
            "affine": affine_matrix,
            "dim_original": dim_original,
            "spacing": self.pixel_spacings[idx],
            "padding": padding,
        }

    def _create_mask(self, landmark: torch.Tensor, dim: tuple[int, ...]) -> torch.Tensor:
        """
        Create a mask from a landmark.

        Args:
            landmark (torch.Tensor): landmark of the image.
            dim (tuple[int, ...]): dimension of the image.

        Returns:
            mask (torch.Tensor): mask.
        """
        mask = torch.zeros((landmark.shape[0], *dim))
        for i in range(landmark.shape[0]):
            if len(landmark.shape) == 3:
                for j in range(landmark.shape[1]):
                    if self.spatial_dims == 2:
                        if 0 <= landmark[i, j, 0] < dim[0] and 0 <= landmark[i, j, 1] < dim[1]:
                            mask[i, int(landmark[i, j, 0]), int(landmark[i, j, 1])] = 1
                    else:
                        if (
                            0 <= landmark[i, j, 0] < dim[0]
                            and 0 <= landmark[i, j, 1] < dim[1]
                            and 0 <= landmark[i, j, 2] < dim[2]
                        ):
                            mask[
                                i,
                                int(landmark[i, j, 0]),
                                int(landmark[i, j, 1]),
                                int(landmark[i, j, 2]),
                            ] = 1
            else:
                if self.spatial_dims == 2:
                    if 0 <= landmark[i, 0] < dim[0] and 0 <= landmark[i, 1] < dim[1]:
                        mask[i, int(landmark[i, 0]), int(landmark[i, 1])] = 1
                else:
                    if (
                        0 <= landmark[i, 0] < dim[0]
                        and 0 <= landmark[i, 1] < dim[1]
                        and 0 <= landmark[i, 2] < dim[2]
                    ):
                        mask[i, int(landmark[i, 0]), int(landmark[i, 1]), int(landmark[i, 2])] = 1
        return mask

    def _read_extract_landmarks(self, paths: list[str]) -> torch.Tensor:
        """
        Read masks and extract landmarks from these masks.

        Args:
            paths (list[str]): list of paths to the masks.

        Returns:
            landmarks (torch.Tensor): landmarks.
        """
        print(f"Reading and extracting masks from {len(paths)} images...")
        landmarks = []
        for path in tqdm(paths):
            landmark = self._read_extract_landmark(path)
            landmarks.append(landmark)
        max_len = 1
        for landmark in landmarks:
            max_len = max(landmark.shape[1], max_len)
        if max_len != 1:
            landmarks = [
                F.pad(landmark, (0, 0, 0, max_len - landmark.shape[1]), value=torch.nan)
                for landmark in landmarks
            ]
        else:
            # Remove sample channel if every class has only one landmark
            landmarks = [landmark.squeeze(0) for landmark in landmarks]
        return torch.stack(landmarks)

    def _read_extract_landmark(self, path: str) -> torch.Tensor:
        """
        Read a mask and extract the landmark from this mask.

        Args:
            path (str): path to the mask.

        Returns:
            landmarks (torch.Tensor): landmark.
        """
        mask = self.mask_loader(path)
        landmarks = self._extract_landmark_from_mask(mask)
        return landmarks

    def _extract_landmark_from_mask(self, mask: torch.Tensor) -> torch.Tensor:
        """
        Extract the landmark from a mask.

        Args:
            mask (torch.Tensor): mask.

        Returns:
            landmarks (torch.Tensor): landmark.
        """
        landmarks_list = []
        max_len = 0
        if self.nb_landmarks == 1:
            landmarks_list.append(mask[0].nonzero().float())
            max_len = max(len(landmarks_list[0].shape), max_len)
        else:
            for i in range(self.nb_landmarks):
                landmarks_list.append((mask[0] == i + 1).nonzero().float())
                max_len = max(len(landmarks_list[i].shape), 0)
        if max_len == 0:
            raise ValueError("No landmark found in mask")
        if max_len == 1 or self.nb_landmarks == 1:
            landmarks = torch.stack(landmarks_list)
        else:
            landmarks_list = [
                landmark.view(self.nb_landmarks, -1, self.spatial_dims) for landmark in landmarks
            ]
            landmarks_list = [
                F.pad(landmark, (0, 0, 0, max_len - landmark.shape[1]), value=torch.nan)
                for landmark in landmarks_list
            ]
            landmarks = torch.cat(landmarks_list)
        return landmarks


class PatchMaskDataset(PatchDataset):
    """
    ``PatchMaskDataset`` is a subclass of ``PatchDataset``. It represents a dataset of images and
    landmarks. The images can be provided as a list of paths to the images or as a list of numpy
    arrays or as a numpy array/torch.Tensor. The landmarks can be provided as a list of paths to
    the masks or as a list of numpy arrays or as a numpy array/torch.Tensor. The masks must be
    grayscale images with the landmarks represented by different values. The landmarks can be
    extracted from the masks and stored as a torch.Tensor or they can be genererated from provided
    landmarks. The images are cropped around the landmarks and stored as torch.Tensor. The landmarks
    are the landmarks of the cropped images. The offset is randomly generated in a range of
    [-range_aug_patch, range_aug_patch] and added to the landmarks to create the cropped images.
    If the cropped images are not of the same size, they are resized to the same size.
    """

    def __init__(
        self,
        imgs: list[str] | list[np.ndarray] | np.ndarray | torch.Tensor,
        landmarks: np.ndarray | torch.Tensor,
        index_landmark=0,
        spatial_dims: int = 2,
        pixel_spacing: Optional[torch.Tensor] = None,
        class_names: Optional[list] = None,
        transform: Optional[Callable] = None,
        store_imgs: bool = False,
        dim_patch: tuple[int, ...] = (256, 256),
        range_aug_patch: int = 64,
    ) -> None:
        super().__init__(
            imgs=imgs,
            landmarks=landmarks,
            index_landmark=index_landmark,
            spatial_dims=spatial_dims,
            pixel_spacing=pixel_spacing,
            class_names=class_names,
            transform=transform,
            store_imgs=store_imgs,
            dim_patch=dim_patch,
            range_aug_patch=range_aug_patch,
        )

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        batch = super().__getitem__(idx)
        mask = self._create_mask(
            batch["landmark"].unsqueeze(0), batch["image"].shape[-self.spatial_dims :]
        )
        return {**batch, "mask": mask}

    def _create_mask(self, landmark: torch.Tensor, dim: tuple[int, ...]) -> torch.Tensor:
        """
        Create a mask from a landmark.

        Args:
            landmark (torch.Tensor): landmark of the image.
            dim (tuple[int, ...]): dimension of the image.

        Returns:
            mask (torch.Tensor): mask.
        """
        mask = torch.zeros((landmark.shape[0], *dim))
        for i in range(landmark.shape[0]):
            if len(landmark.shape) == 3:
                for j in range(landmark.shape[1]):
                    if self.spatial_dims == 2:
                        if 0 <= landmark[i, j, 0] < dim[0] and 0 <= landmark[i, j, 1] < dim[1]:
                            mask[i, int(landmark[i, j, 0]), int(landmark[i, j, 1])] = 1
                    else:
                        if (
                            landmark[i, j, 0] < dim[0]
                            and landmark[i, j, 1] < dim[1]
                            and landmark[i, j, 2] < dim[2]
                        ):
                            mask[
                                i,
                                int(landmark[i, j, 0]),
                                int(landmark[i, j, 1]),
                                int(landmark[i, j, 2]),
                            ] = 1
            else:
                if self.spatial_dims == 2:
                    if 0 <= landmark[i, 0] < dim[0] and 0 <= landmark[i, 1] < dim[1]:
                        mask[i, int(landmark[i, 0]), int(landmark[i, 1])] = 1
                else:
                    if (
                        landmark[i, 0] < dim[0]
                        and landmark[i, 1] < dim[1]
                        and landmark[i, 2] < dim[2]
                    ):
                        mask[i, int(landmark[i, 0]), int(landmark[i, 1]), int(landmark[i, 2])] = 1
        return mask


class HeatmapDataset(LandmarkDataset):
    """
    ``HeatmapDataset`` is a subclass of ``LandmarkDataset``. It represents a dataset of images and
    landmarks. The images can be provided as a list of paths to the images or as a list of numpy
    arrays or as a numpy array/torch.Tensor. Th landmarks must be provided. The heatmaps are created
    from the landmarks using a Gaussian or Laplacian function. When using the ``HeatmapDataset``,
    images are always stored in memory.

    Args:
        imgs (list[str] | list[np.array] | np.ndarray | torch.Tensor): list of paths to the images
            or list of numpy arrays or numpy array/torch.Tensor.
        landmarks (np.ndarray | torch.Tensor): landmarks of the images. If the landmarks are 2D: y, x
            or 3D: z, y, x.
        spatial_dims (int): number of spatial dimensions of the images. (defaults: 2)
        pixel_spacing (Optional[np.ndarray | torch.Tensor]): pixel spacing of the images.
            (defaults: None)
        class_names (Optional[list]): names of the landmarks. (defaults: None)
        transform (Optional[Callable]): transformation to apply to the images and landmarks.
            (defaults: None)
        sigma (float | list[float]): sigma of the Gaussian or Laplacian function. (defaults: 5)
        dim_img (Optional[tuple[int, int]]): dimension of the images. (defaults: None)
        batch_size (int): batch size to use when creating the heatmaps. (defaults: 32)
        background (bool): whether to add a background channel or not. (defaults: False)
        all_points (bool): whether to add a channel for all points or not. (defaults: False)
        gamma (Optional[float]): gamma of the Gaussian or Laplacian function. (defaults: None)
        heatmap_fun (str): function to use to create the heatmaps. (defaults: "gaussian")
        heatmap_size (Optional[tuple[int, int]]): size of the heatmaps. (defaults: None)
        img_paths (Optional[list[str]]): list of paths to the images. (defaults: None)
        resize_pad (bool): whether to resize the images and landmarks or not. (defaults: True)
    """

    def __init__(
        self,
        imgs: list[str] | list[np.ndarray] | np.ndarray | torch.Tensor,
        landmarks: np.ndarray | torch.Tensor,
        spatial_dims: int = 2,
        pixel_spacing: Optional[np.ndarray | torch.Tensor] = None,
        class_names: Optional[list] = None,
        transform: Optional[Callable] = None,
        store_imgs: bool = True,
        sigma: float | list[float] | np.ndarray | torch.Tensor = 5,
        dim_img: Optional[tuple[int, ...]] = None,
        batch_size: int = 32,
        background: bool = False,
        all_points: bool = False,
        gamma: Optional[float] = None,
        heatmap_fun: str = "gaussian",
        heatmap_size: Optional[tuple[int, ...]] = None,
        img_paths: Optional[list[str]] = None,
        resize_pad: bool = True,
    ) -> None:
        super().__init__(
            imgs=imgs,
            landmarks=landmarks,
            spatial_dims=spatial_dims,
            pixel_spacing=pixel_spacing,
            class_names=class_names,
            transform=transform,
            store_imgs=store_imgs,
            dim_img=dim_img,
            img_paths=img_paths,
            resize_pad=resize_pad,
        )
        if heatmap_size is None:
            if self.dim_img is not None:
                heatmap_size = self.dim_img
            else:
                assert (
                    self.store_imgs
                ), "Heatmap size must be provided if imgs are not stored, and dim_img is None"
                # heatmap size becomes the size of the first original image
                if self.spatial_dims == 2:
                    heatmap_size = (
                        int(self.dim_original[0, 0].item()),
                        int(self.dim_original[0, 1].item()),
                    )
                else:
                    heatmap_size = (
                        int(self.dim_original[0, 0].item()),
                        int(self.dim_original[0, 1].item()),
                        int(self.dim_original[0, 2].item()),
                    )
        self.background = background
        self.all_points = all_points
        self.heatmap_generator: HeatmapGenerator
        if heatmap_fun == "gaussian":
            self.heatmap_generator = GaussianHeatmapGenerator(
                nb_landmarks=len(self.class_names),
                sigmas=sigma,
                gamma=gamma,
                heatmap_size=heatmap_size,
                learnable=False,
                background=background,
                all_points=all_points,
            )
        elif heatmap_fun == "laplacian":
            self.heatmap_generator = LaplacianHeatmapGenerator(
                nb_landmarks=len(self.class_names),
                sigmas=sigma,
                gamma=gamma,
                heatmap_size=heatmap_size,
                learnable=False,
                background=background,
                all_points=all_points,
            )
        else:
            raise ValueError("Heatmap function not supported")
        if batch_size > 1:
            self.batch_size = batch_size
        else:
            self.batch_size = 2
        if self.store_imgs:
            self.heatmaps = self._create_heatmaps_batch(self.landmarks)

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        if self.store_imgs:
            img = self.imgs[idx]
            heatmap = self.heatmaps[idx]
            landmark = self.landmarks[idx]
            padding = self.paddings[idx]
            dim_original = self.dim_original[idx]
        else:
            if not hasattr(self, "image_loader"):
                img = self.imgs[idx]  # type: ignore
            else:
                img = self.image_loader(self.img_paths[idx])  # type: ignore
            dim_original = torch.tensor(img.shape[-self.spatial_dims :]).float()
            if self.dim_img is not None:
                if self.resize_pad:
                    dim_original = torch.tensor(img.shape[-self.spatial_dims :]).float()
                    img, pad = resize_with_pad(img, self.dim_img)
                    padding = torch.Tensor(pad)
                    landmark = resize_landmarks(
                        self.landmarks_original[idx], dim_original, self.dim_img, padding
                    )
                else:
                    img = Resize(self.dim_img)(img)
                    landmark = resize_landmarks(
                        self.landmarks_original[idx], dim_original, self.dim_img
                    )
                    padding = torch.zeros(2)
            heatmap = self._create_heatmap(landmark).squeeze(0)
        if self.transform is not None:
            (
                img,
                heatmap,
                landmark_t,
                affine_matrix,
            ) = self._transform_img_heatmap_landmark(img, heatmap, landmark, self.transform)
        else:
            landmark_t = landmark
            affine_matrix = torch.eye(4)
        return {
            "image": img,
            "mask": heatmap,
            "landmark": landmark_t,
            "affine": affine_matrix,
            "dim_original": dim_original,
            "spacing": self.pixel_spacings[idx],
            "padding": padding,
        }

    def _transform_img_heatmap_landmark(
        self, img: torch.Tensor, heatmap: torch.Tensor, landmark: torch.Tensor, transform: Callable
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Transform the image, the heatmap and the landmark using the transformation
        ``self.transform``.

        Args:
            img (torch.Tensor): image to transform.
            heatmap (torch.Tensor): heatmap to transform.
            landmark (torch.Tensor): landmark to transform.

        Returns:
            img_t (torch.Tensor): transformed image.
            heatmap_t (torch.Tensor): transformed heatmap.
            landmark_t (torch.Tensor): transformed landmark.
            affine_matrix_push (torch.Tensor): push affine matrix of the transformation.
        """
        out = transform({"image": img, "mask": heatmap.squeeze(0)})
        img_t = out["image"]
        heatmap_t = out["mask"]
        affine_matrix_pull = img_t.meta.get("affine", None).float()
        affine_matrix_push = torch.linalg.inv(affine_matrix_pull)
        landmark_t = affine_landmarks(landmark, affine_matrix_push)
        return img_t, heatmap_t, landmark_t, affine_matrix_push

    def _create_heatmaps(self, landmarks: torch.Tensor) -> list[torch.Tensor]:
        """
        Create heatmaps from landmarks.

        Args:
            landmarks (torch.Tensor): landmarks of the images.

        Returns:
            heatmaps (list[torch.Tensor]): list of heatmaps.
        """
        print(f"Creating heatmaps for {len(landmarks)} images...")
        heatmaps = []
        for i in tqdm(range(landmarks.shape[0])):
            heatmaps.append(self._create_heatmap(landmarks[i]))
        return heatmaps

    def _create_heatmaps_batch(self, landmarks: torch.Tensor) -> torch.Tensor:
        """
        Create heatmaps from landmarks using batches.

        Args:
            landmarks (torch.Tensor): landmarks of the images.

        Returns:
            heatmaps (torch.Tensor): heatmaps.
        """
        print(f"Creating heatmaps for {len(landmarks)} images... (batch size: {self.batch_size})")
        heatmaps = torch.zeros(
            (
                len(landmarks),
                len(self.class_names) + int(self.all_points or self.background),
                self.heatmap_generator.heatmap_size[0],
                self.heatmap_generator.heatmap_size[1],
            )
        )
        i = 0
        for i in tqdm(range(landmarks.shape[0] // self.batch_size)):
            heatmaps[i * self.batch_size : (i + 1) * self.batch_size] = self.heatmap_generator(
                landmarks[i * self.batch_size : (i + 1) * self.batch_size]
            )
        if landmarks.shape[0] % self.batch_size != 0:
            try:
                heatmaps[(i + 1) * self.batch_size :] = self.heatmap_generator(
                    landmarks[(i + 1) * self.batch_size :]
                )
            except RuntimeError:
                heatmaps = self.heatmap_generator(landmarks)
        return heatmaps

    def _create_heatmap(self, landmark: torch.Tensor) -> torch.Tensor:
        """
        Create a heatmap from a landmark.

        Args:
            landmark (torch.Tensor): landmark of the image.

        Returns:
            heatmap (torch.Tensor): heatmap.
        """
        return self.heatmap_generator(landmark.unsqueeze(0))
